# (c) 2024 Niels Provos
#
# This file contains functions for creating and exporting glTF files.
# We generate a glTF file representing a 3D scene with a camera, cards, and image slices.
# The resulting file can be opened in a 3D application like Blender, Houdini or Unreal.
#
import base64

import numpy as np
import pygltflib as gltf
from PIL import Image


def rotation_quaternion_y(y_rot_degrees):
    """Calculates the rotation quaternion for a rotation around the y-axis.

    Args:
        y_rot_degrees: The rotation angle in degrees.

    Returns:
        A NumPy array representing the rotation quaternion (x, y, z, w).
    """

    # Convert to radians and half the angle
    theta = np.radians(y_rot_degrees) / 2
    axis = np.array([0, 1, 0])  # Rotation around the y-axis

    quaternion = np.zeros(4)
    quaternion[:3] = axis * np.sin(theta)
    quaternion[3] = np.cos(theta)

    return quaternion.tolist()


def create_camera(
    gltf_obj, focal_length, aspect_ratio, translation, rotation_quarternion
):
    """
    Creates a camera in the glTF object with the specified parameters.

    Args:
        gltf_obj (gltf.Gltf): The glTF object to add the camera to.
        focal_length (float): The focal length of the camera.
        aspect_ratio (float): The aspect ratio of the camera.
        translation (List[float]): The translation of the camera node.
        rotation_quarternion (List[float]): The rotation of the camera node as a quaternion.

    Returns:
        int: The index of the created camera.

    """
    camera_index = len(gltf_obj.cameras)

    sensor_width = 35.0  # Sensor width in mm
    sensor_height = sensor_width / aspect_ratio

    # Create the camera object
    camera = gltf.Camera(
        type="perspective",
        name=f"Camera_{camera_index}",
        perspective=gltf.Perspective(
            aspectRatio=aspect_ratio,
            yfov=2 * np.arctan(sensor_height / focal_length),
            znear=0.01,
            zfar=10000,
        ),
    )
    gltf_obj.cameras.append(camera)

    # Create the camera node
    camera_node = gltf.Node(
        translation=translation, rotation=rotation_quarternion, camera=camera_index
    )
    gltf_obj.nodes.append(camera_node)

    return camera_index


def create_buffer_and_view(gltf_obj, data, target=gltf.ARRAY_BUFFER):
    """
    Creates a buffer and buffer view in a glTF object.

    Args:
        gltf_obj (gltf.Gltf): The glTF object to add the buffer and buffer view to.
        data (numpy.ndarray): The data to be stored in the buffer.
        target (int): The target usage of the buffer view (default: gltf.ARRAY_BUFFER).

    Returns:
        int: the index of the created buffer view.
    """
    tmp_buffer = gltf.Buffer(
        byteLength=data.nbytes,
        uri=f"data:application/octet-stream;base64,{base64.b64encode(data.tobytes()).decode()}",
    )
    gltf_obj.buffers.append(tmp_buffer)
    tmp_buffer_index = len(gltf_obj.buffers) - 1

    tmp_buffer_view = gltf.BufferView(
        buffer=tmp_buffer_index, byteOffset=0, byteLength=data.nbytes, target=target
    )
    gltf_obj.bufferViews.append(tmp_buffer_view)
    tmp_buffer_view_index = len(gltf_obj.bufferViews) - 1

    return tmp_buffer_view_index


def subdivide_geometry(coords, subdivisions, dimension):
    """
    Subdivides a plane into a grid with the specified number of subdivisions. Can handle both 3D and 2D geometries.

    Args:
        coords (numpy.ndarray): The corner coordinates of the geometry (3D for spatial coordinates, 2D for texture coordinates).
        subdivisions (int): The number of subdivisions to create.
        dimension (int): The dimension of the target points (3 for spatial coordinates, 2 for texture coordinates).

    Returns:
        numpy.ndarray: The corner coordinates of the subdivided geometry.
    """
    x = np.linspace(coords[0, 0], coords[1, 0], subdivisions + 1, dtype=np.float32)
    y = np.linspace(coords[0, 1], coords[3, 1], subdivisions + 1, dtype=np.float32)
    x, y = np.meshgrid(x, y)

    if dimension == 3:
        z = np.zeros_like(x)
        points = np.stack([x, y, z], axis=-1)
    elif dimension == 2:
        points = np.stack([x, y], axis=-1)

    return points.reshape(-1, dimension)


def triangle_indices_from_grid(vertices):
    """
    Generates triangle indices for a grid of vertices.

    Args:
        vertices (numpy.ndarray): The 3D corner coordinates of the grid.

    Returns:
        numpy.ndarray: The triangle indices for the grid.
    """
    # Calculate the number of vertices in each row
    row_length = int(np.sqrt(len(vertices)))

    # Create the indices for the triangles
    indices = []
    for i in range(row_length - 1):
        for j in range(row_length - 1):
            # Calculate the indices for the current quad
            tl = i * row_length + j
            tr = tl + 1
            bl = (i + 1) * row_length + j
            br = bl + 1

            # Create the two triangles for the quad
            indices.append([tl, tr, bl])
            indices.append([bl, tr, br])

    return np.array(indices, dtype=np.uint32)


def displace_vertices(vertices, depth_map, displacement_scale=10.0):
    """
    Displaces the vertices of a plane based on a depth map.

    Args:
        vertices (numpy.ndarray): The 3D corner coordinates of the plane.
        depth_map (numpy.ndarray): The depth map to displace the vertices with. Normalized to [0, 1].

    Returns:
        numpy.ndarray: The displaced vertices.
    """
    # Get the dimensions of the depth map
    depth_map_width, depth_map_height = depth_map.shape

    # Calculate the texture coordinates for the vertices
    tex_coords = vertices[:, :2].copy()
    tex_coords[:, 1] = -tex_coords[:, 1]  # flip Y-axis

    # normalize the texture coordinates to [0, 1]
    tex_min_x, tex_min_y = tex_coords.min(axis=0)
    tex_max_x, tex_max_y = tex_coords.max(axis=0)
    tex_coords -= [tex_min_x, tex_min_y]
    tex_coords /= [tex_max_x - tex_min_x, tex_max_y - tex_min_y]

    # Calculate the pixel coordinates for the texture coordinates
    pixel_coords = (tex_coords * [depth_map_height - 1, depth_map_width - 1]).astype(
        int
    )

    # Get the depth values for the pixel coordinates
    depths = depth_map[pixel_coords[:, 1], pixel_coords[:, 0]] * displacement_scale

    # Displace the vertices based on the depth values
    vertices[:, 2] = depths

    return vertices


def create_card(
    gltf_obj, i, corners_3d, subdivisions=300, depth_map=None, displacement_scale=0.0
):
    """
    Creates a card (plane) in the glTF object with the specified parameters.

    Args:
        gltf_obj (gltf.Gltf): The glTF object to add the card to.
        i (int): The index of the card.
        corners_3d (numpy.ndarray): The 3D corner coordinates for the card.
        subdivisions (int, optional): The number of subdivisions for the card. Defaults to 300.
        depth_map (numpy.ndarray, optional): The depth map for the card. Defaults to None.
        displacement_scale (float, optional): The scale of the displacement. Defaults to 0.0.

    Returns:
        int: The index of the created mesh.
    """
    # Set the vertices and indices for the plane

    # negate the y coordinates of corners_3d
    vertices = np.array(corners_3d, dtype=np.float32)
    vertices[:, 1] = -vertices[:, 1]

    # reorder the vertices of the 4 point plane
    tl = vertices[0]
    tr = vertices[1]
    bl = vertices[3]
    br = vertices[2]

    vertices = np.array([tl, tr, bl, br], dtype=np.float32)

    tex_coords = np.array([[0, 0], [1, 0], [0, 1], [1, 1]], dtype=np.float32)

    if displacement_scale > 0.0 and depth_map is not None:
        vertices = subdivide_geometry(vertices, subdivisions, 3)
        vertices = displace_vertices(
            vertices, depth_map, displacement_scale=displacement_scale
        )
        tex_coords = subdivide_geometry(tex_coords, subdivisions, 2)

    indices = triangle_indices_from_grid(vertices)

    # Create the buffer and buffer view for vertices
    vertex_bufferview_index = create_buffer_and_view(
        gltf_obj, vertices, target=gltf.ARRAY_BUFFER
    )

    # Create the buffer and buffer view for texture coordinates
    tex_coord_bufferview_index = create_buffer_and_view(
        gltf_obj, tex_coords, target=gltf.ARRAY_BUFFER
    )

    # Create the buffer and buffer view for indices
    index_bufferview_index = create_buffer_and_view(
        gltf_obj, indices, target=gltf.ELEMENT_ARRAY_BUFFER
    )

    # Create the accessor for texture coordinates
    tex_coord_accessor = gltf.Accessor(
        bufferView=tex_coord_bufferview_index,
        componentType=gltf.FLOAT,
        count=len(tex_coords),
        type=gltf.VEC2,
        max=tex_coords.max(axis=0).tolist(),
        min=tex_coords.min(axis=0).tolist(),
    )
    gltf_obj.accessors.append(tex_coord_accessor)
    tex_coord_accessor_index = len(gltf_obj.accessors) - 1

    # Create the accessor for vertices
    vertex_accessor = gltf.Accessor(
        bufferView=vertex_bufferview_index,
        componentType=gltf.FLOAT,
        count=len(vertices),
        type=gltf.VEC3,
        max=vertices.max(axis=0).tolist(),
        min=vertices.min(axis=0).tolist(),
    )
    gltf_obj.accessors.append(vertex_accessor)
    vertex_accessor_index = len(gltf_obj.accessors) - 1

    # Create the accessor for indices
    index_accessor = gltf.Accessor(
        bufferView=index_bufferview_index,
        componentType=gltf.UNSIGNED_INT,
        count=indices.size,
        type=gltf.SCALAR,
    )
    gltf_obj.accessors.append(index_accessor)
    index_accessor_index = len(gltf_obj.accessors) - 1

    card_name = f"Card_{i}"

    # Create the mesh for the plane
    mesh = gltf.Mesh(
        name=card_name,
        primitives=[
            gltf.Primitive(
                attributes=gltf.Attributes(
                    POSITION=vertex_accessor_index,
                    TEXCOORD_0=tex_coord_accessor_index,
                ),
                indices=index_accessor_index,
                material=i,
            )
        ],
    )

    return mesh


def export_gltf(
    output_path,
    cam,
    image_slices,
    image_paths,
    depth_paths=[],
    displacement_scale=0.0,
    inline_images=True,
    support_dof=False,
):
    """
    Export the camera, cards, and image slices to a glTF file.

    Args:
        output_path (str): The path to save the glTF file.
        aspect_ratio (float): The aspect ratio of the camera.
        focal_length (float): The focal length of the camera.
        camera_distance (float): The distance of the camera from the origin.
        cam (Camera): The camera object for the scene.
        image_slices (list): List of 3D corner coordinates for each card.
        image_paths (list): List of file paths for each image slice.
        depth_paths (list, optional): List of file paths for each depth map. Defaults to [].
        displacement_scale (float, optional): The scale of the displacement. Defaults to 0.0.
        inline_images (bool, optional): Whether to inline the images in the glTF file. Defaults to True.
    """

    # compute pre-requisites
    image_height, image_width = image_slices[0].image.shape[:2]
    camera_matrix = cam.camera_matrix(image_width, image_height)
    aspect_ratio = float(camera_matrix[0, 2]) / camera_matrix[1, 2]
    focal_length = cam.focal_length
    camera_distance = cam.camera_distance

    # Create a new glTF object
    gltf_obj = gltf.GLTF2(scene=0)

    # Create the scene
    scene = gltf.Scene()
    gltf_obj.scenes.append(scene)

    camera_index = create_camera(
        gltf_obj,
        focal_length,
        aspect_ratio,
        [0, 0, -camera_distance],
        rotation_quaternion_y(180),
    )
    # Add the camera node to the scene
    scene.nodes.append(camera_index)

    subdivisions = 500

    alpha_mode = "MASK" if support_dof else "BLEND"

    # Create the card objects (planes)
    for i, image_slice in enumerate(image_slices):
        corners_3d = image_slice.create_card(image_height, image_width, cam)
        # Translaton hack so that we can put the depth on the node
        z_transform = corners_3d[0][2]
        corners_3d[:, 2] -= z_transform

        depth_map = None
        if len(depth_paths) > i:
            depth_map = Image.open(depth_paths[i])
            width, height = depth_map.size
            depth_map = depth_map.resize(
                (subdivisions + 1, subdivisions + 1), Image.BICUBIC
            )
            depth_map = depth_map.resize((width, height), Image.BICUBIC)
            depth_map = np.array(depth_map)
            depth_map = depth_map.astype(np.float32) / 255.0

        mesh = create_card(
            gltf_obj,
            i,
            corners_3d,
            subdivisions,
            depth_map,
            displacement_scale=displacement_scale,
        )
        gltf_obj.meshes.append(mesh)

        # Create the material and assign the texture
        material = gltf.Material(
            name=f"Material_{i}",
            pbrMetallicRoughness=gltf.PbrMetallicRoughness(
                baseColorTexture=gltf.TextureInfo(index=i)
            ),
            # Set the emissive color (RGB values)
            emissiveFactor=[1.0, 1.0, 1.0],
            emissiveTexture=gltf.TextureInfo(index=i),
            alphaMode=alpha_mode,
            alphaCutoff=0.5 if alpha_mode == "MASK" else None,
            doubleSided=True,
        )

        image = gltf.Image(uri=str(image_paths[i]))
        gltf_obj.images.append(image)

        texture = gltf.Texture(
            source=i,
        )
        gltf_obj.textures.append(texture)

        gltf_obj.materials.append(material)

        # Create the card node and add it to the scene
        card_node = gltf.Node(
            mesh=i,
            translation=[0, 0, int(z_transform)],
            rotation=rotation_quaternion_y(180),
        )
        gltf_obj.nodes.append(card_node)
        scene.nodes.append(len(gltf_obj.nodes) - 1)

    # Save the glTF file
    if inline_images:
        gltf_obj.convert_images(gltf.ImageFormat.DATAURI)
    else:
        gltf_obj.convert_images(gltf.ImageFormat.FILE)

    gltf_obj.save(str(output_path))

    return str(output_path)



#!/usr/bin/env python3
"""
Setup script for Parallax Maker development and packaging.
"""

import os
import sys
import subprocess
import shutil
from pathlib import Path


def run_command(cmd, description):
    """Run a command and handle errors."""
    print(f"Running: {description}")
    print(f"Command: {cmd}")
    result = subprocess.run(cmd, shell=True, capture_output=True, text=True)
    if result.returncode != 0:
        print(f"Error: {result.stderr}")
        return False
    print(f"Success: {result.stdout}")
    return True


def clean_build():
    """Clean previous build artifacts."""
    print("Cleaning previous build artifacts...")
    dirs_to_remove = ["dist", "build", "*.egg-info"]
    for pattern in dirs_to_remove:
        for path in Path(".").glob(pattern):
            if path.is_dir():
                shutil.rmtree(path)
                print(f"Removed directory: {path}")
            else:
                path.unlink()
                print(f"Removed file: {path}")


def build_package():
    """Build the package."""
    print("Building package...")
    return run_command("python -m build", "Building wheel and source distribution")


def check_package():
    """Check the built package."""
    print("Checking package...")
    return run_command("python -m twine check dist/*", "Checking package integrity")


def install_dev_tools():
    """Install development tools."""
    print("Installing development tools...")
    tools = ["build", "twine", "pytest", "black", "flake8"]
    for tool in tools:
        if not run_command(f"pip install {tool}", f"Installing {tool}"):
            return False
    return True


def run_tests():
    """Run tests."""
    print("Running tests...")
    return run_command("python -m pytest", "Running test suite")


def format_code():
    """Format code with black."""
    print("Formatting code...")
    return run_command("python -m black .", "Formatting code with black")


def main():
    """Main setup function."""
    if len(sys.argv) < 2:
        print("Usage: python setup_dev.py <command>")
        print("Commands:")
        print("  install-tools  - Install development tools")
        print("  clean         - Clean build artifacts")
        print("  build         - Build the package")
        print("  check         - Check the package")
        print("  test          - Run tests")
        print("  format        - Format code")
        print("  full-build    - Clean, build, and check")
        print("  dev-setup     - Full development setup")
        sys.exit(1)

    command = sys.argv[1]

    if command == "install-tools":
        install_dev_tools()
    elif command == "clean":
        clean_build()
    elif command == "build":
        build_package()
    elif command == "check":
        check_package()
    elif command == "test":
        run_tests()
    elif command == "format":
        format_code()
    elif command == "full-build":
        clean_build()
        if build_package():
            check_package()
    elif command == "dev-setup":
        install_dev_tools()
        format_code()
        run_tests()
        clean_build()
        if build_package():
            check_package()
            print("\n✅ Development setup complete!")
            print("📦 Package built successfully!")
            print("\nNext steps:")
            print("1. Test install: pip install dist/*.whl")
            print(
                "2. Upload to TestPyPI: python -m twine upload --repository testpypi dist/*"
            )
            print("3. Upload to PyPI: python -m twine upload dist/*")
    else:
        print(f"Unknown command: {command}")
        sys.exit(1)


if __name__ == "__main__":
    main()



import unittest
from unittest.mock import patch, MagicMock

import io
import base64
from PIL import Image

import dash
from dash import html

from . import constants as C
import numpy as np
from .components import (
    make_inpainting_container_callbacks,
    make_inpainting_container,
    make_configuration_container,
)
from .slice import ImageSlice


class TestUpdateInpaintingImageDisplay(unittest.TestCase):
    def setUp(self):
        # Setup a minimal Dash app
        self.app = dash.Dash(__name__)
        self.app.layout = html.Div(
            [
                make_inpainting_container(),
                make_configuration_container(),
            ]
        )

        # Save the callback
        self._update_inpainting_image_display = make_inpainting_container_callbacks(
            self.app
        )

    @patch("parallax_maker.components.ctx", new_callable=MagicMock)
    @patch("parallax_maker.components.Image.open")
    @patch("parallax_maker.components.Image.fromarray")
    @patch("parallax_maker.components.Path")
    @patch("parallax_maker.controller.AppState.from_cache")
    @patch("parallax_maker.inpainting.InpaintingModel")
    def test_callback_triggered_comfyui(
        self,
        mock_model,
        mock_from_cache,
        mock_path,
        mock_image_open,
        mock_image_fromarray,
        mock_callback_context,
    ):  # Mock callback context
        mock_callback_context.triggered_id = C.BTN_GENERATE_INPAINTING

        # Setup test data
        n_clicks = 1
        filename = "test_filename"
        model = "comfyui"
        server_address = "http://localhost:8000"
        workflow = "data:filetype;base64,workflow_data"
        positive_prompt = "sky is clear"
        negative_prompt = "cloudy sky"
        strength = 0.7
        guidance_scale = 5.0
        padding = 10
        blur = 5

        # Workflow path mock
        workflow_path = MagicMock()
        workflow_path.exists.return_value = True

        # Set up the AppState mock
        state = MagicMock()
        state.selected_slice = (
            0  # Ensure this matches the state expected by the function
        )
        state.mask_filename.return_value = "mask_0.png"
        state.image_slices = [ImageSlice(np.zeros((100, 100, 4)))]
        state.workflow_path.return_value = workflow_path
        mock_from_cache.return_value = state

        # Mock mask Path exists
        mock_path.return_value.exists.return_value = True

        # Return a mock Image when open succeeds
        mock_img = Image.new("L", (100, 100))
        mock_image_open.return_value = mock_img

        # Setup InpaintingModel mock
        pipeline = MagicMock()
        mock_model.return_value = pipeline
        pipeline.inpaint.return_value = "painted_image"  # Mock the result of inpainting

        # Return the final image
        mock_image_fromarray.return_value = Image.new("RGBA", (100, 100))

        # Call the function under test
        results = self._update_inpainting_image_display(
            n_clicks,
            None,
            None,
            filename,
            model,
            workflow,
            positive_prompt,
            negative_prompt,
            strength,
            guidance_scale,
            padding,
            blur,
        )

        # Verify outputs and behaviors
        pipeline.load_model.assert_called_once()  # Check that model was loaded
        self.assertEqual(len(results[0]), 3)  # Assuming 3 images are expected
        for i, img in enumerate(results[0]):
            id = img.id
            self.assertDictEqual(id, {"type": C.ID_INPAINTING_IMAGE, "index": i})
            img_data = img.src
            # Assuming you have some format to represent the image
            img = Image.open(io.BytesIO(base64.b64decode(img_data.split(",")[1])))
            self.assertEqual(img.size, (100, 100))

        # make sure the second return is a list
        self.assertIsInstance(results[1], list)
        # If this is expected to be an empty list
        self.assertEqual(len(results[1]), 0)

    @patch("parallax_maker.components.ctx", new_callable=MagicMock)
    @patch("parallax_maker.components.Image.open")
    @patch("parallax_maker.components.Image.fromarray")
    @patch("parallax_maker.components.Path")
    @patch("parallax_maker.controller.AppState.from_cache")
    @patch("parallax_maker.inpainting.InpaintingModel")
    def test_callback_triggered_normal(
        self,
        mock_model,
        mock_from_cache,
        mock_path,
        mock_image_open,
        mock_image_fromarray,
        mock_callback_context,
    ):
        # Mock callback context
        mock_callback_context.triggered_id = C.BTN_GENERATE_INPAINTING

        # Setup test data
        n_clicks = 1
        filename = "test_filename"
        model = "other_model"
        server_address = None
        workflow = None
        positive_prompt = "sky is clear"
        negative_prompt = "cloudy sky"
        strength = 0.7
        guidance_scale = 5.0
        padding = 10
        blur = 5

        # Workflow path mock
        workflow_path = MagicMock()
        workflow_path.exists.return_value = False

        # Set up the AppState mock
        state = MagicMock()
        state.selected_slice = (
            0  # Ensure this matches the state expected by the function
        )
        state.mask_filename.return_value = "mask_0.png"
        state.image_slices = [ImageSlice(np.zeros((100, 100, 4)))]
        state.workflow_path.return_value = workflow_path
        mock_from_cache.return_value = state

        # Mock mask Path exists
        mock_path.return_value.exists.return_value = True

        # Return a mock Image when open succeeds
        mock_img = Image.new("L", (100, 100))
        mock_image_open.return_value = mock_img

        # Setup InpaintingModel mock
        pipeline = MagicMock()
        mock_model.return_value = pipeline
        pipeline.inpaint.return_value = "painted_image"  # Mock the result of inpainting

        # Return the final image
        mock_image_fromarray.return_value = Image.new("RGBA", (100, 100))

        # Call the function under test
        results = self._update_inpainting_image_display(
            n_clicks,
            None,
            None,
            filename,
            model,
            workflow,
            positive_prompt,
            negative_prompt,
            strength,
            guidance_scale,
            padding,
            blur,
        )

        # Verify outputs and behaviors
        pipeline.load_model.assert_called_once()  # Check that model was loaded
        self.assertEqual(len(results[0]), 3)  # Assuming 3 images are expected
        for i, img in enumerate(results[0]):
            id = img.id
            self.assertDictEqual(id, {"type": C.ID_INPAINTING_IMAGE, "index": i})
            img_data = img.src
            # Assuming you have some format to represent the image
            img = Image.open(io.BytesIO(base64.b64decode(img_data.split(",")[1])))
            self.assertEqual(img.size, (100, 100))

        # make sure the second return is a list
        self.assertIsInstance(results[1], list)
        # If this is expected to be an empty list
        self.assertEqual(len(results[1]), 0)


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
# (c) 2024 Niels Provos

import base64
import io
import os
from pathlib import Path
from PIL import Image
import numpy as np

from . import constants as C
from .segmentation import (
    generate_depth_map,
    analyze_depth_histogram,
    generate_image_slices,
    create_slice_from_mask,
    export_gltf,
    blend_with_alpha,
    remove_mask_from_alpha,
    render_image_sequence,
)
from . import components
from .utils import (
    find_pixel_from_event,
    postprocess_depth_map,
    get_gltf_iframe,
    get_no_gltf_available,
    to_data_url,
)
from .depth import DepthEstimationModel
from .instance import SegmentationModel
from .inpainting import InpaintingModel, create_inpainting_pipeline
from .clientside import make_clientside_callbacks
from .slice import ImageSlice


import dash
from dash import dcc, html, ctx, no_update
from dash.dependencies import Input, Output, State
from dash.dependencies import ALL, MATCH
from dash_extensions import EventListener
from dash.exceptions import PreventUpdate
from flask import send_file
from werkzeug import serving

from .controller import AppState, CompositeMode
from .camera import Camera

# Globals
EXPAND_MASK = 5
HIGHLIGHT_COLOR = "color-is-selected-light"

# Progress tracking variables
current_progress = -1
total_progress = 100


def progress_callback(current, total):
    global current_progress, total_progress
    current_progress = (current / total) * 100
    total_progress = 100


# call the ability to add external scripts
external_scripts = [
    # add the tailwind cdn url hosting the files with the utility classes
    {"src": "https://kit.fontawesome.com/48f728cfc9.js"},
]

app = dash.Dash(__name__, external_scripts=external_scripts)

# Create a Flask route for serving images


@app.server.route(f"/{AppState.SRV_DIR}/<path:filename>")
def serve_data(filename):
    filename = Path(os.getcwd()) / filename
    if filename.suffix == ".gltf":
        mimetype = "model/gltf+json"
    else:
        mimetype = f"image/{filename.suffix[1:]}"
    return send_file(str(filename), mimetype=mimetype)


# JavaScript event(s) that we want to listen to and what properties to collect.
eventScroll = {"event": "scroll", "props": ["type", "scrollLeft", "scrollTop"]}

app.layout = html.Div(
    [
        EventListener(events=[eventScroll], logging=True, id="evScroll"),
        # dcc.Store stores all application state
        dcc.Store(id=C.STORE_APPSTATE_FILENAME),
        dcc.Store(id=C.STORE_RESTORE_STATE),  # trigger to restore state
        # Store for rect coordinates from the clientside JS
        dcc.Store(id=C.STORE_RECT_DATA),
        # Store for bounding box coordinates from inpainting
        dcc.Store(id=C.STORE_BOUNDING_BOX),
        dcc.Store(id=C.LOGS_DATA, data=[]),  # Store for logs
        # Trigger for generating depth map
        dcc.Store(id=C.STORE_TRIGGER_GEN_DEPTHMAP),
        # Trigger for updating depth map
        dcc.Store(id=C.STORE_TRIGGER_UPDATE_DEPTHMAP),
        # Trigger for updating thresholds
        dcc.Store(id=C.STORE_UPDATE_THRESHOLD_CONTAINER),
        # Context for the help window
        dcc.Store(id=C.STORE_CURRENT_TAB),
        # App Layout
        html.Header(
            [
                html.Div(
                    [
                        html.Button(
                            [html.I(className="fas fa-moon", id=C.ICON_DARK_MODE)],
                            id=C.BTN_DARK_MODE,
                            n_clicks=0,
                            className="dark-mode-toggle",
                        )
                    ],
                    className="header-left",
                ),
                html.H1("Parallax Maker", className="title-text"),
                html.Div([], className="header-right"),
            ],
            className="title-header",
        ),
        html.Main(
            [
                html.Div(
                    ["Some helpful text to guide the user"],
                    id=C.CTR_HELP_WINDOW,
                    className="help-box hidden absolute w-64 p-2 z-40",
                ),
                components.make_tabs(
                    "viewer",
                    ["2D", "3D"],
                    [
                        components.make_input_image_container(
                            upload_id=C.UPLOAD_IMAGE,
                            image_id=C.IMAGE,
                            event_id="el",
                            canvas_id=C.CANVAS,
                            outer_class_name="w-full col-span-3",
                        ),
                        html.Div(
                            id=C.CTR_MODEL_VIEWER,
                            children=[
                                html.Iframe(
                                    id=C.IFRAME_MODEL_VIEWER,
                                    srcDoc=get_no_gltf_available(),
                                    style={"height": "70vh"},
                                    className="gltf-container",
                                )
                            ],
                        ),
                    ],
                    outer_class_name="w-full col-span-3",
                ),
                components.make_tabs(
                    "main",
                    ["Mode", "Segmentation", "Inpainting", "Export", "Configuration"],
                    [
                        html.Div(
                            [
                                components.make_depth_map_container(
                                    depth_map_id=C.CTR_DEPTH_MAP
                                ),
                                components.make_mode_selector(),
                            ],
                            className="w-full",
                            id="depth-map-column",
                        ),
                        components.make_slice_generation_container(),
                        components.make_inpainting_container(),
                        html.Div(
                            [
                                components.make_3d_export_div(),
                                components.make_animation_export_div(),
                            ],
                            className="w-full",
                        ),
                        components.make_configuration_div(),
                    ],
                    outer_class_name="w-full col-span-2",
                ),
            ],
            className="grid grid-cols-5 gap-4 p-2",
        ),
        components.make_logs_container(logs_id="log"),
        html.Footer("© 2024 Niels Provos", className="footer"),
    ],
    id="app-container",
    className="min-h-screen",
)

app.scripts.config.serve_locally = True

make_clientside_callbacks(app)

components.make_segmentation_callbacks(app)
components.make_canvas_callbacks(app)
components.make_navigation_callbacks(app)
components.make_inpainting_container_callbacks(app)
components.make_configuration_callbacks(app)


# Callbacks for collapsible sections
components.make_tabs_callback(app, "viewer")
components.make_tabs_callback(app, "main")
components.make_tools_callbacks(app)


@app.callback(
    Output("app-container", "className"),
    Output(C.ICON_DARK_MODE, "className"),
    Input(C.BTN_DARK_MODE, "n_clicks"),
    State(C.STORE_APPSTATE_FILENAME, "data"),
    prevent_initial_call=True,
)
def toggle_dark_mode(n_clicks, filename):

    dark_mode = n_clicks % 2 == 1
    if filename is not None:
        state = AppState.from_cache(filename)
        if state.dark_mode != dark_mode:
            state.dark_mode = dark_mode
            state.to_file(
                filename,
                save_image_slices=False,
                save_depth_map=False,
                save_input_image=False,
            )

    if dark_mode:
        return "dark min-h-screen", "fas fa-sun"
    else:
        return "min-h-screen", "fas fa-moon"


# Callback for the logs


@app.callback(
    Output("log", "children"), Input(C.LOGS_DATA, "data"), prevent_initial_call=True
)
def update_logs(data):
    structured_logs = [html.Div(log) for log in data[-3:]]
    return structured_logs


# Callback to update progress bar


@app.callback(
    Output(C.CTR_PROGRESS_BAR, "children"),
    Output(C.PROGRESS_INTERVAL, "disabled", allow_duplicate=True),
    Input(C.PROGRESS_INTERVAL, "n_intervals"),
    prevent_initial_call=True,
)
def update_progress(n):
    progress_bar = html.Div(
        className="progress-bar-fill", style={"width": f"{max(0, current_progress)}%"}
    )
    interval_disabled = current_progress >= total_progress or current_progress == -1
    return progress_bar, interval_disabled


@app.callback(
    Output({"type": "threshold-slider", "index": ALL}, "value"),
    Output(C.IMAGE, "src", allow_duplicate=True),
    Input({"type": "threshold-slider", "index": ALL}, "value"),
    State(C.SLIDER_NUM_SLICES, "value"),
    State(C.STORE_APPSTATE_FILENAME, "data"),
    prevent_initial_call=True,
)
def update_threshold_values(threshold_values, num_slices, filename):
    if filename is None:
        raise PreventUpdate()

    state = AppState.from_cache(filename)

    if state.imgThresholds[1:-1] == threshold_values:
        print("Threshold values are the same; not erasing data.")
        raise PreventUpdate()

    # make sure that threshold values are monotonically increasing
    if threshold_values[0] <= 0:
        threshold_values[0] = 1

    for i in range(1, num_slices - 1):
        if threshold_values[i] <= threshold_values[i - 1]:
            threshold_values[i] = threshold_values[i - 1] + 1

    # go through the list in reverse order to make sure that the thresholds are monotonically decreasing
    if threshold_values[-1] >= 255:
        threshold_values[-1] = 254

    # num slices is the number of thresholds + 1, so the largest index is num_slices - 2
    # and the second largest index is num_slices - 3
    for i in range(num_slices - 3, -1, -1):
        if threshold_values[i] >= threshold_values[i + 1]:
            threshold_values[i] = threshold_values[i + 1] - 1

    state.imgThresholds[1:-1] = threshold_values

    img_data = no_update
    if state.slice_pixel:
        state.slice_mask, _ = state.depth_slice_from_pixel(
            state.slice_pixel[0], state.slice_pixel[1]
        )
        if state.slice_mask is not None:
            result = state.apply_mask(state.imgData, state.slice_mask)
            img_data = state.serve_main_image(result)
        else:
            img_data = state.serve_main_image(state.imgData)

    return threshold_values, img_data


@app.callback(
    Output(C.CTR_THRESHOLDS, "children"),
    Input(C.STORE_UPDATE_THRESHOLD_CONTAINER, "data"),
    State(C.STORE_APPSTATE_FILENAME, "data"),
    prevent_initial_call=True,
)
def update_thresholds_html(value, filename):
    if filename is None:
        raise PreventUpdate()

    state = AppState.from_cache(filename)
    thresholds = []
    for i in range(1, state.num_slices):
        threshold = html.Div(
            [
                dcc.Slider(
                    id={"type": "threshold-slider", "index": i},
                    min=0,
                    max=255,
                    step=1,
                    value=state.imgThresholds[i],
                    marks=None,
                    tooltip={"always_visible": True, "placement": "right"},
                )
            ],
            className="m-1 pl-1",
        )
        thresholds.append(threshold)

    return thresholds


@app.callback(
    Output(C.STORE_UPDATE_THRESHOLD_CONTAINER, "data", allow_duplicate=True),
    Output(C.LOGS_DATA, "data", allow_duplicate=True),
    # triggers regeneration of slices if we have them already
    Input(C.CTR_DEPTH_MAP, "children"),
    Input(C.SLIDER_NUM_SLICES, "value"),
    State(C.STORE_APPSTATE_FILENAME, "data"),
    State(C.LOGS_DATA, "data"),
    prevent_initial_call=True,
)
def update_thresholds(contents, num_slices, filename, logs_data):
    if filename is None:
        raise PreventUpdate()

    state = AppState.from_cache(filename)
    if (
        state.num_slices == num_slices
        and state.imgThresholds is not None
        and len(state.imgThresholds) == num_slices + 1
    ):
        print("Number of slices is the same; not erasing data.")
        raise PreventUpdate()

    state.num_slices = num_slices

    if state.depthMapData is None:
        logs_data.append("No depth map data available")
        state.imgThresholds = [0]
        state.imgThresholds.extend(
            [i * (255 // (num_slices - 1)) for i in range(1, num_slices)]
        )
    elif state.imgThresholds is None or len(state.imgThresholds) != num_slices:
        state.imgThresholds = analyze_depth_histogram(
            state.depthMapData, num_slices=num_slices
        )

    logs_data.append(f"Thresholds: {state.imgThresholds}")

    return True, logs_data


@app.callback(
    Output(C.STORE_APPSTATE_FILENAME, "data", allow_duplicate=True),
    Output(C.STORE_UPDATE_SLICE, "data", allow_duplicate=True),
    Output(C.STORE_TRIGGER_GEN_DEPTHMAP, "data", allow_duplicate=True),
    Output(C.IMAGE, "src", allow_duplicate=True),
    Output(C.CTR_DEPTH_MAP, "children", allow_duplicate=True),
    Output(C.PROGRESS_INTERVAL, "disabled", allow_duplicate=True),
    Input(C.UPLOAD_IMAGE, "contents"),
    State({"type": f"tab-content-main", "index": ALL}, "className"),
    prevent_initial_call=True,
)
def update_input_image(contents, classnames):
    if not contents:
        raise PreventUpdate()

    # allow an upload action only when the user is on the main tab
    # or configuration tab
    on_valid_tab = "hidden" not in classnames[0] or "hidden" not in classnames[-1]
    if classnames is None or not on_valid_tab:
        raise PreventUpdate()

    state, filename = AppState.from_file_or_new(None)

    content_type, content_string = contents.split(",")

    # save the image data to the state
    state.set_img_data(Image.open(io.BytesIO(base64.b64decode(content_string))))

    img_uri = state.serve_input_image()

    return (
        filename,
        True,
        True,
        img_uri,
        html.Img(id="depthmap-image", className="w-full p-0 object-scale-down"),
        False,
    )


@app.callback(
    Output(C.IMAGE, "src", allow_duplicate=True),
    Output(C.LOGS_DATA, "data"),
    Output(C.LOADING_UPLOAD, "children", allow_duplicate=True),
    Output(C.STORE_CLICKED_POINT, "data"),
    Input(C.SEG_MULTI_COMMIT, "n_clicks"),
    Input("el", "n_events"),
    State("el", "event"),
    State(C.STORE_RECT_DATA, "data"),
    State(C.DROPDOWN_MODE_SELECTOR, "value"),
    State(C.STORE_APPSTATE_FILENAME, "data"),
    State(C.LOGS_DATA, "data"),
    prevent_initial_call=True,
)
def click_event(n_clicks, n_events, e, rect_data, mode, filename, logs_data):
    if filename is None:
        raise PreventUpdate()

    state = AppState.from_cache(filename)

    t_id = ctx.triggered_id
    shiftClick = False
    ctrlClick = False

    if t_id == "el":
        if e is None or rect_data is None or state.imgData is None:
            raise PreventUpdate()

        pixel_x, pixel_y = find_pixel_from_event(state, e, rect_data)

        # we need to find the depth even if we use instance segmentation
        new_mask, depth = state.depth_slice_from_pixel(pixel_x, pixel_y)
        state.slice_pixel = (pixel_x, pixel_y)
        state.slice_pixel_depth = depth

        logs_data.append(
            f"Click event at pixel coordinates ({pixel_x}, {pixel_y}) at depth {depth}"
        )

        shiftClick = e["shiftKey"]
        ctrlClick = e["ctrlKey"]
    elif t_id != C.SEG_MULTI_COMMIT:
        raise ValueError(f"Unexpected trigger {t_id}")

    image = state.imgData
    if mode == "segment":
        positive_points = []
        negative_points = []

        if state.multi_point_mode:
            # if we are still selecting points, add them to the state
            if t_id != C.SEG_MULTI_COMMIT:
                state.points_selected.append(((pixel_x, pixel_y), ctrlClick))
                return no_update, no_update, no_update, e
            # if we are committing the points, add them to the positive and negative points
            for point, ctrl_click in state.points_selected:
                if ctrl_click:
                    negative_points.append(point)
                else:
                    positive_points.append(point)
        else:
            assert t_id == "el"
            positive_points.append((pixel_x, pixel_y))

        if state.segmentation_model == None:
            state.segmentation_model = SegmentationModel()
        # if we have a slice, take it and compose the background image over it
        if state.selected_slice is not None:
            image = state.slice_image_composed(state.selected_slice, CompositeMode.NONE)
        state.segmentation_model.segment_image(image)
        # XXX - allow selection of the cheap vs the expensive alogrithm
        new_mask = state.segmentation_model.mask_at_point_blended(
            {"positive_points": positive_points, "negative_points": negative_points}
        )

        logs_data.append(
            f"Committed points {positive_points} and {negative_points} for Segment Anything"
        )

    # allow mask manipulation with add and subtract via shift and ctrl click
    if state.slice_mask is None or not (shiftClick or ctrlClick):
        state.slice_mask = new_mask
    elif shiftClick:
        state.slice_mask = np.maximum(state.slice_mask, new_mask)
    elif ctrlClick:
        state.slice_mask = np.minimum(state.slice_mask, 255 - new_mask)

    if state.slice_mask is not None:
        result = state.apply_mask(image, state.slice_mask)
        img_data = state.serve_main_image(result)
    else:
        img_data = state.serve_main_image(state.imgData)

    return img_data, logs_data, "", no_update


@app.callback(
    Output(C.STORE_TRIGGER_GEN_DEPTHMAP, "data"),
    Input(C.BTN_GENERATE_DEPTHMAP, "n_clicks"),
    State(C.STORE_APPSTATE_FILENAME, "data"),
    running=[(Output(C.BTN_GENERATE_DEPTHMAP, "disabled"), True, False)],
    prevent_initial_call=True,
)
def generate_depth_map_from_button(n_clicks, filename):
    if n_clicks is None or filename is None:
        raise PreventUpdate()
    return True


@app.callback(
    Output(C.STORE_TRIGGER_UPDATE_DEPTHMAP, "data"),
    Output(C.DEPTHMAP_OUTPUT, "children"),
    Input(C.STORE_TRIGGER_GEN_DEPTHMAP, "data"),
    State(C.STORE_APPSTATE_FILENAME, "data"),
    State(C.DROPDOWN_DEPTH_MODEL, "value"),
    prevent_initial_call=True,
)
def generate_depth_map_callback(ignored_data, filename, model):
    if filename is None:
        raise PreventUpdate()

    print(f"Received a request to generate a depth map for state f{filename}")
    state = AppState.from_cache(filename)

    PIL_image = state.imgData

    if PIL_image.mode == "RGBA":
        PIL_image = PIL_image.convert("RGB")

    np_image = np.array(PIL_image)

    depth_model = DepthEstimationModel(model=model)
    if depth_model != state.depth_estimation_model:
        state.depth_estimation_model = depth_model

    state.depthMapData = generate_depth_map(
        np_image,
        model=state.depth_estimation_model,
        progress_callback=progress_callback,
    )
    state.imgThresholds = None

    state.to_file(
        filename, save_image_slices=False, save_depth_map=True, save_input_image=False
    )

    return True, ""


@app.callback(
    Output(C.CTR_DEPTH_MAP, "children"),
    Input(C.STORE_TRIGGER_UPDATE_DEPTHMAP, "data"),
    State(C.STORE_APPSTATE_FILENAME, "data"),
    prevent_initial_call=True,
)
def update_depth_map_callback(ignored_data, filename):
    if filename is None:
        raise PreventUpdate()

    state = AppState.from_cache(filename)
    depth_map_pil = Image.fromarray(state.depthMapData)

    buffered = io.BytesIO()
    depth_map_pil.save(buffered, format="PNG")
    img_str = base64.b64encode(buffered.getvalue()).decode("utf-8")

    return (
        html.Img(
            src="data:image/png;base64,{}".format(img_str),
            className="w-full h-full object-contain",
            style={"height": "35vh"},
            id="depthmap-image",
        ),
        "",
    )


@app.callback(
    Output(C.IMAGE, "src", allow_duplicate=True),
    Output(C.STORE_UPDATE_SLICE, "data", allow_duplicate=True),
    Output(C.LOGS_DATA, "data", allow_duplicate=True),
    Output(C.LOADING_UPLOAD, "children", allow_duplicate=True),
    Input(C.BTN_DELETE_SLICE, "n_clicks"),
    State(C.STORE_APPSTATE_FILENAME, "data"),
    State(C.LOGS_DATA, "data"),
    prevent_initial_call=True,
)
def delete_slice_request(n_clicks, filename, logs):
    if n_clicks is None:
        raise PreventUpdate()

    if filename is None:
        raise PreventUpdate()

    state = AppState.from_cache(filename)
    if state.selected_slice is None:
        logs.append("No slice selected")
        return no_update, no_update, logs, ""

    logs.append(f"Deleted slice at index {state.selected_slice}")

    state.delete_slice(state.selected_slice)

    # sufficient to just change the json.
    state.to_file(
        filename, save_image_slices=False, save_depth_map=False, save_input_image=False
    )

    return state.serve_main_image(state.imgData), True, logs, ""


@app.callback(
    Output(C.LOGS_DATA, "data", allow_duplicate=True),
    Input(C.BTN_COPY_SLICE, "n_clicks"),
    State(C.STORE_APPSTATE_FILENAME, "data"),
    State(C.LOGS_DATA, "data"),
    prevent_initial_call=True,
)
def copy_to_clipboard(n_clicks, filename, logs):
    if n_clicks is None:
        raise PreventUpdate()

    if filename is None:
        raise PreventUpdate()

    state = AppState.from_cache(filename)
    if state.slice_mask is None:
        logs.append("No mask selected")
        return logs

    if state.selected_slice is not None:
        image = state.slice_image_composed(state.selected_slice, CompositeMode.NONE)
    else:
        image = state.imgData
    image = np.array(image.convert("RGBA"))
    image[:, :, 3] = state.slice_mask

    state.clipboard_image = image

    logs.append("Copied mask to clipboard")
    return logs


@app.callback(
    Output(C.STORE_UPDATE_SLICE, "data", allow_duplicate=True),
    Output(C.LOGS_DATA, "data", allow_duplicate=True),
    Output(C.LOADING_UPLOAD, "children", allow_duplicate=True),
    Input(C.BTN_PASTE_SLICE, "n_clicks"),
    State(C.STORE_APPSTATE_FILENAME, "data"),
    State(C.LOGS_DATA, "data"),
    prevent_initial_call=True,
)
def paste_clipboard_request(n_clicks, filename, logs):
    if n_clicks is None:
        raise PreventUpdate()

    if filename is None:
        raise PreventUpdate()

    state = AppState.from_cache(filename)
    if state.clipboard_image is None:
        logs.append("Nothing in the clipboard")
        return no_update, logs, no_update

    if state.selected_slice is None:
        logs.append("No slice selected")
        return no_update, logs, no_update

    image = state.clipboard_image
    # updates the image slice in place - dangerous
    blend_with_alpha(state.image_slices[state.selected_slice].image, image)
    state.image_slices[state.selected_slice].new_version()

    state.to_file(
        filename, save_image_slices=False, save_depth_map=False, save_input_image=False
    )

    logs.append(f"Pasted clipboard to slice {state.selected_slice}")
    return True, logs, ""


@app.callback(
    Output(C.STORE_UPDATE_SLICE, "data", allow_duplicate=True),
    Output(C.LOGS_DATA, "data", allow_duplicate=True),
    Output(C.LOADING_UPLOAD, "children", allow_duplicate=True),
    Input(C.BTN_REMOVE_SLICE, "n_clicks"),
    State(C.STORE_APPSTATE_FILENAME, "data"),
    State(C.LOGS_DATA, "data"),
    prevent_initial_call=True,
)
def remove_mask_slice_request(n_clicks, filename, logs):
    if n_clicks is None:
        raise PreventUpdate()

    if filename is None:
        raise PreventUpdate()

    state = AppState.from_cache(filename)
    if state.slice_mask is None:
        logs.append("No mask selected")
        return no_update, logs, no_update

    if state.selected_slice is None:
        logs.append("No slice selected")
        return no_update, logs, no_update

    final_mask = remove_mask_from_alpha(
        state.image_slices[state.selected_slice].image, state.slice_mask
    )
    state.image_slices[state.selected_slice].image[:, :, 3] = final_mask
    state.image_slices[state.selected_slice].new_version()

    state.to_file(
        filename, save_image_slices=False, save_depth_map=False, save_input_image=False
    )

    logs.append(f"Removed mask from slice {state.selected_slice}")
    return True, logs, ""


@app.callback(
    Output(C.STORE_UPDATE_SLICE, "data", allow_duplicate=True),
    Output(C.LOGS_DATA, "data", allow_duplicate=True),
    Output(C.LOADING_UPLOAD, "children", allow_duplicate=True),
    Input(C.BTN_ADD_SLICE, "n_clicks"),
    State(C.STORE_APPSTATE_FILENAME, "data"),
    State(C.LOGS_DATA, "data"),
    running=[(Output(C.BTN_ADD_SLICE, "disabled"), True, False)],
    prevent_initial_call=True,
)
def add_mask_slice_request(n_clicks, filename, logs):
    if n_clicks is None:
        raise PreventUpdate()

    if filename is None:
        raise PreventUpdate()

    state = AppState.from_cache(filename)
    if state.slice_mask is None:
        logs.append("No mask selected")
        return no_update, logs, no_update

    if state.selected_slice is None:
        logs.append("No slice selected")
        return no_update, logs, no_update

    # XXX - should we create an option to copy from the composed image?
    image = create_slice_from_mask(
        state.imgData, state.slice_mask, num_expand=EXPAND_MASK
    )
    # updates the image slice in place - dangerous
    blend_with_alpha(state.image_slices[state.selected_slice].image, image)
    state.image_slices[state.selected_slice].new_version()

    state.to_file(
        filename, save_image_slices=False, save_depth_map=False, save_input_image=False
    )

    logs.append(f"Added mask to slice {state.selected_slice}")
    return True, logs, ""


@app.callback(
    Input(C.TEXT_POSITIVE_PROMPT, "value"),
    Input(C.TEXT_NEGATIVE_PROMPT, "value"),
    State(C.STORE_APPSTATE_FILENAME, "data"),
    prevent_initial_call=True,
)
def update_prompt_text(positive, negative, filename):
    if filename is None:
        raise PreventUpdate()

    state = AppState.from_cache(filename)
    if state.selected_slice is None:
        raise PreventUpdate()

    if (
        state.image_slices[state.selected_slice].positive_prompt == positive
        and state.image_slices[state.selected_slice].negative_prompt == negative
    ):
        raise PreventUpdate()

    state.image_slices[state.selected_slice].positive_prompt = positive
    state.image_slices[state.selected_slice].negative_prompt = negative

    state.to_file(
        filename, save_image_slices=False, save_depth_map=False, save_input_image=False
    )


@app.callback(
    Output(C.STORE_UPDATE_SLICE, "data", allow_duplicate=True),
    Output(C.TEXT_POSITIVE_PROMPT, "value", allow_duplicate=True),
    Output(C.TEXT_NEGATIVE_PROMPT, "value", allow_duplicate=True),
    Output(C.LOGS_DATA, "data", allow_duplicate=True),
    Output(C.LOADING_UPLOAD, "children", allow_duplicate=True),
    Input(C.BTN_CREATE_SLICE, "n_clicks"),
    State(C.STORE_APPSTATE_FILENAME, "data"),
    State(C.LOGS_DATA, "data"),
    running=[(Output(C.BTN_CREATE_SLICE, "disabled"), True, False)],
    prevent_initial_call=True,
)
def create_single_slice_request(n_clicks, filename, logs):
    if n_clicks is None:
        raise PreventUpdate()

    if filename is None:
        raise PreventUpdate()

    state = AppState.from_cache(filename)
    if state.imgData is None:
        logs.append("No image available")
        return no_update, no_update, no_update, logs, no_update

    depth = 127
    if state.slice_mask is None:
        # create an empty image that the user can inpaint if they want to
        image = Image.new("RGBA", state.imgData.size, (0, 0, 0, 0))
    else:
        if state.slice_pixel is not None:
            depth = state.slice_pixel_depth
        image = create_slice_from_mask(
            state.imgData, state.slice_mask, num_expand=EXPAND_MASK
        )

    image = ImageSlice(image, depth)
    state.selected_slice = state.add_slice(image)
    image.save_image()
    state.to_file(
        filename, save_image_slices=False, save_depth_map=False, save_input_image=False
    )

    logs.append("Created a slice from the mask")

    return True, "", "", logs, ""


@app.callback(
    Output(C.STORE_UPDATE_SLICE, "data", allow_duplicate=True),
    Output(C.LOGS_DATA, "data", allow_duplicate=True),
    Input(C.BTN_BALANCE_SLICE, "n_clicks"),
    State(C.STORE_APPSTATE_FILENAME, "data"),
    State(C.LOGS_DATA, "data"),
    prevent_initial_call=True,
)
def balance_slices_request(n_clicks, filename, logs):
    if n_clicks is None:
        raise PreventUpdate()

    if filename is None:
        raise PreventUpdate()

    state = AppState.from_cache(filename)
    if len(state.image_depths) == 0:
        raise PreventUpdate()

    state.balance_slices_depths()
    state.to_file(
        filename, save_image_slices=False, save_depth_map=False, save_input_image=False
    )

    logs.append("Balanced slice depths")

    return True, logs


@app.callback(
    Output(C.STORE_GENERATE_SLICE, "data"), Input(C.BTN_GENERATE_SLICE, "n_clicks")
)
def generate_slices_request(n_clicks):
    if n_clicks is None:
        raise PreventUpdate()
    return n_clicks


@app.callback(
    Output(C.CTR_SLICE_IMAGES, "children"),
    Output("gen-slice-output", "children", allow_duplicate=True),
    Output(C.IMAGE, "src", allow_duplicate=True),
    Input(C.STORE_UPDATE_SLICE, "data"),
    State(C.STORE_APPSTATE_FILENAME, "data"),
    prevent_initial_call=True,
)
def update_slices(ignored_data, filename):
    if filename is None:
        raise PreventUpdate()

    state = AppState.from_cache(filename)
    if len(state.image_slices) == 0:
        # a user may have uploaded a new image and not generated slices yet
        return [], "", no_update

    if state.depthMapData is None:
        raise PreventUpdate()

    caret_color_enabled = "has-history-color"
    caret_color_disabled = "no-history-color"

    img_container = []
    for i, image_slice in enumerate(state.image_slices):
        img_data = state.serve_slice_image(i)

        left_color = (
            caret_color_enabled
            if image_slice.can_undo(forward=False)
            else caret_color_disabled
        )
        left_disabled = True if left_color == caret_color_disabled else False
        right_color = (
            caret_color_enabled
            if image_slice.can_undo(forward=True)
            else caret_color_disabled
        )
        right_disabled = True if right_color == caret_color_disabled else False

        left_id = {"type": "slice-undo-backwards", "index": i}
        right_id = {"type": "slice-undo-forwards", "index": i}

        slice_name = html.Div(
            [
                html.Button(
                    title="Download image for manipuation in an external editor",
                    className="fa-solid fa-download pr-1",
                    id={"type": "slice-info", "index": i},
                ),
                html.Button(
                    title="Undo last change",
                    className=f"fa-solid fa-caret-left {left_color} pr-1",
                    id=left_id,
                    disabled=left_disabled,
                ),
                html.Button(
                    title="Redo last change",
                    className=f"fa-solid fa-caret-right {right_color} pr-1",
                    id=right_id,
                    disabled=right_disabled,
                ),
                Path(image_slice.filename).stem,
            ]
        )

        # slice creation with select a slice so we need to highlight it here
        highlight_class = f"overlay" if state.selected_slice == i else "hidden"
        img_container.append(
            dcc.Upload(
                html.Div(
                    [
                        html.Div(
                            # The number to display
                            children=f"{int(image_slice.depth)}",
                            id={"type": C.ID_SLICE_DEPTH_DISPLAY, "index": i},
                            className="depth-number-display",
                        ),
                        html.Div(
                            className=highlight_class,
                            id={"type": C.ID_SLICE_OVERLAY, "index": i},
                        ),
                        dcc.Input(
                            id={"type": C.INPUT_SLICE_DEPTH, "index": i},
                            className="depth-number-input hidden",
                            type="number",
                            debounce=True,
                            inputMode="numeric",
                            maxLength=3,
                            value=f"{int(image_slice.depth)}",
                        ),
                        html.Img(
                            src=img_data,
                            className="image-border",
                            id={"type": "slice", "index": i},
                        ),
                        html.Div(
                            children=slice_name,
                            className="text-center text-overlay p-1",
                        ),
                    ],
                    style={"position": "relative"},
                ),
                id={"type": C.UPLOAD_SLICE, "index": i},
                disable_click=True,
            )
        )

    img_data = no_update
    if state.selected_slice is not None:
        assert state.selected_slice >= 0 and state.selected_slice < len(
            state.image_slices
        )
        mode = (
            CompositeMode.CHECKERBOARD
            if state.use_checkerboard
            else CompositeMode.GRAYSCALE
        )
        img_data = state.serve_slice_image_composed(state.selected_slice, mode=mode)
        state.slice_pixel = None
        state.slice_pixel_depth = None
        state.slice_mask = None

    return img_container, "", img_data


@app.callback(
    Output({"type": C.INPUT_SLICE_DEPTH, "index": MATCH}, "className"),
    Input({"type": C.ID_SLICE_DEPTH_DISPLAY, "index": MATCH}, "n_clicks"),
    State({"type": C.INPUT_SLICE_DEPTH, "index": MATCH}, "className"),
    prevent_initial_call=True,
)
def display_depth_input(n_clicks, class_name):
    if n_clicks is None:
        raise PreventUpdate()

    class_name = class_name.replace("hidden", "")
    return class_name


@app.callback(
    Output(C.STORE_UPDATE_SLICE, "data", allow_duplicate=True),
    Output(C.STORE_INPAINTING, "data", allow_duplicate=True),
    Input({"type": C.INPUT_SLICE_DEPTH, "index": ALL}, "value"),
    Input({"type": C.INPUT_SLICE_DEPTH, "index": ALL}, "n_submit"),
    State(C.STORE_APPSTATE_FILENAME, "data"),
    prevent_initial_call=True,
)
def record_depth_input(values, n_submits, filename):
    if filename is None or values is None or n_submits is None:
        raise PreventUpdate()

    index = ctx.triggered_id["index"]
    if n_submits[index] is None:
        raise PreventUpdate()

    value = int(values[index])

    # need to re-order and validate the depth values
    state = AppState.from_cache(filename)

    new_index = state.change_slice_depth(index, value)
    if index != new_index:
        state.selected_slice = None

    state.to_file(
        filename, save_image_slices=False, save_depth_map=False, save_input_image=False
    )
    return True, True


@app.callback(
    Output(C.STORE_UPDATE_SLICE, "data", allow_duplicate=True),
    Input({"type": "slice-undo-backwards", "index": ALL}, "n_clicks"),
    Input({"type": "slice-undo-forwards", "index": ALL}, "n_clicks"),
    State(C.STORE_APPSTATE_FILENAME, "data"),
    prevent_initial_call=True,
)
def undo_slice(n_clicks_backwards, n_clicks_forwards, filename):
    if filename is None:
        raise PreventUpdate()

    state = AppState.from_cache(filename)

    # don't need to use ctx.triggered_id since we are repaining the whole thing
    index = None
    forward = None
    if any(n_clicks_backwards):
        index = n_clicks_backwards.index(1)
        forward = False
    elif any(n_clicks_forwards):
        index = n_clicks_forwards.index(1)
        forward = True
    else:
        raise PreventUpdate()

    if not state.image_slices[index].undo(forward=forward):
        print(f"Cannot undo slice {index} with forward {forward}")
        raise PreventUpdate()

    # only save the json with the updated file mapping
    state.to_file(
        filename, save_image_slices=False, save_depth_map=False, save_input_image=False
    )

    return True


@app.callback(
    Output(C.STORE_UPDATE_SLICE, "data", allow_duplicate=True),
    Output("gen-slice-output", "children"),
    Input(C.STORE_GENERATE_SLICE, "data"),
    State(C.STORE_APPSTATE_FILENAME, "data"),
    prevent_initial_call=True,
)
def generate_slices(ignored_data, filename):
    if filename is None:
        raise PreventUpdate()

    state = AppState.from_cache(filename)
    if state.depthMapData is None:
        raise PreventUpdate()

    # XXX - refactor the state update into the AppState class
    state.image_slices = generate_image_slices(
        np.array(state.imgData),
        state.depthMapData,
        state.imgThresholds,
        num_expand=EXPAND_MASK,
    )

    print(f"Generated {len(state.image_slices)} image slices; saving to file")
    state.to_file(filename)

    return True, ""


@app.callback(
    Output(C.IMAGE, "src"),
    Output({"type": C.ID_SLICE_OVERLAY, "index": ALL}, "className"),
    Output(C.TEXT_POSITIVE_PROMPT, "value", allow_duplicate=True),
    Output(C.TEXT_NEGATIVE_PROMPT, "value", allow_duplicate=True),
    Output(C.STORE_INPAINTING, "data", allow_duplicate=True),
    Input({"type": "slice", "index": ALL}, "n_clicks"),
    Input({"type": C.ID_SLICE_OVERLAY, "index": ALL}, "n_clicks"),
    State({"type": "slice", "index": ALL}, "id"),
    State({"type": "slice", "index": ALL}, "src"),
    State({"type": C.ID_SLICE_OVERLAY, "index": ALL}, "className"),
    State(C.STORE_APPSTATE_FILENAME, "data"),
    prevent_initial_call=True,
)
def display_slice(n_clicks, n_clicks_two, id, src, classnames, filename):
    if (n_clicks is None or any(n_clicks) is False) and (
        n_clicks_two is None or any(n_clicks_two) is False
    ):
        raise PreventUpdate()

    if filename is None:
        raise PreventUpdate()

    state = AppState.from_cache(filename)

    index = ctx.triggered_id["index"]

    # if we are already displaying the slice, then we should remove it
    positive_prompt = ""
    negative_prompt = ""
    if state.selected_slice != index:
        state.selected_slice = index
        mode = (
            CompositeMode.CHECKERBOARD
            if state.use_checkerboard
            else CompositeMode.GRAYSCALE
        )
        result = state.serve_slice_image_composed(index, mode=mode)
        positive_prompt = state.image_slices[index].positive_prompt
        negative_prompt = state.image_slices[index].negative_prompt
    else:
        state.selected_slice = None
        result = state.serve_input_image()

    for i in range(len(classnames)):
        classnames[i] = "overlay" if i == state.selected_slice else "hidden"

    return result, classnames, positive_prompt, negative_prompt, True


@app.callback(
    Output(C.LOGS_DATA, "data", allow_duplicate=True),
    Output(C.LOADING_GLTF, "children", allow_duplicate=True),
    Input(C.BTN_UPSCALE_TEXTURES, "n_clicks"),
    State(C.STORE_APPSTATE_FILENAME, "data"),
    State(C.DROPDOWN_INPAINT_MODEL, "value"),
    State(C.INPUT_EXTERNAL_SERVER, "value"),
    State(C.UPLOAD_COMFYUI_WORKFLOW, "contents"),
    State(C.LOGS_DATA, "data"),
    running=[(Output(C.BTN_UPSCALE_TEXTURES, "disabled"), True, False)],
    prevent_initial_call=True,
)
def upscale_texture(n_clicks, filename, model_name, server_address, workflow, logs):
    if filename is None:
        raise PreventUpdate()

    state = AppState.from_cache(filename)

    # this create the pipeline in state and will be used by upscale slices
    create_inpainting_pipeline(model_name, workflow, state)

    state.upscale_slices()

    logs.append("Upscaled textures for slices")

    return logs, ""


@app.callback(
    Output(C.DOWNLOAD_GLTF, "data"),
    Output(C.LOADING_GLTF, "children", allow_duplicate=True),
    Input(C.BTN_GLTF_EXPORT, "n_clicks"),
    State(C.STORE_APPSTATE_FILENAME, "data"),
    State(C.SLIDER_DISPLACEMENT, "value"),
    State(C.CHECKLIST_DOF, "value"),
    running=[(Output(C.BTN_GLTF_EXPORT, "disabled"), True, False)],
    prevent_initial_call=True,
)
def gltf_export(n_clicks, filename, displacement_scale, dof):
    if n_clicks is None or filename is None:
        raise PreventUpdate()

    state = AppState.from_cache(filename)

    gltf_path = export_state_as_gltf(
        state, filename, state.camera, displacement_scale, support_dof=("dof" in dof)
    )

    return dcc.send_file(gltf_path, filename="scene.gltf"), ""


@app.callback(
    Input(C.DROPDOWN_DEPTH_MODEL, "value"),
    State(C.STORE_APPSTATE_FILENAME, "data"),
    prevent_initial_call=True,
)
def remember_depth_model(value, filename):
    if filename is None:
        raise PreventUpdate()

    state = AppState.from_cache(filename)
    if state.depth_model_name == value:
        raise PreventUpdate()

    if state.depth_model_name == value:
        raise PreventUpdate()
    state.depth_model_name = value
    state.to_file(
        filename, save_image_slices=False, save_depth_map=False, save_input_image=False
    )
    return


@app.callback(
    Input(C.SLIDER_CAMERA_DISTANCE, "value"),
    Input(C.SLIDER_FOCAL_LENGTH, "value"),
    Input(C.SLIDER_MAX_DISTANCE, "value"),
    Input(C.SLIDER_DISPLACEMENT, "value"),
    State(C.STORE_APPSTATE_FILENAME, "data"),
    prevent_initial_call=True,
)
def remember_camera_parameters(
    camera_distance, focal_length, max_distance, displacement, filename
):
    if filename is None:
        raise PreventUpdate()

    state = AppState.from_cache(filename)

    # don't save if the values are the same
    camera = Camera(camera_distance, focal_length, max_distance)
    if state.camera == camera and state.mesh_displacement == displacement:
        raise PreventUpdate()

    state.camera.camera_distance = camera_distance
    state.camera.focal_length = focal_length
    state.camera.max_distance = max_distance
    state.mesh_displacement = displacement
    state.to_file(
        filename, save_image_slices=False, save_depth_map=False, save_input_image=False
    )
    return


@app.callback(
    Input(C.DROPDOWN_INPAINT_MODEL, "value"),
    State(C.STORE_APPSTATE_FILENAME, "data"),
    prevent_initial_call=True,
)
def remember_inpaint_model(value, filename):
    if filename is None:
        raise PreventUpdate()

    state = AppState.from_cache(filename)
    if state.inpainting_model_name == value:
        raise PreventUpdate()

    state.inpainting_model_name = value
    state.to_file(
        filename, save_image_slices=False, save_depth_map=False, save_input_image=False
    )
    return


# XXX - this and the callback above can be chained to avoid code duplication


@app.callback(
    Output("model-viewer", "srcDoc", allow_duplicate=True),
    Output(C.LOADING_GLTF, "children", allow_duplicate=True),
    Input(C.BTN_GLTF_CREATE, "n_clicks"),
    State(C.STORE_APPSTATE_FILENAME, "data"),
    State(C.SLIDER_DISPLACEMENT, "value"),
    State(C.DROPDOWN_DEPTH_MODEL, "value"),
    State(C.CHECKLIST_DOF, "value"),
    running=[(Output(C.BTN_GLTF_CREATE, "disabled"), True, False)],
    prevent_initial_call=True,
)
def gltf_create(n_clicks, filename, displacement_scale, model_name, dof):
    if n_clicks is None or filename is None:
        raise PreventUpdate()

    state = AppState.from_cache(filename)

    export_state_as_gltf(
        state,
        filename,
        state.camera,
        displacement_scale,
        modelname=model_name,
        support_dof=("dof" in dof),
    )

    return get_gltf_iframe(state.serve_model_file()), ""


def export_state_as_gltf(
    state: AppState,
    filename,
    camera,
    displacement_scale,
    modelname="midas",
    inline_images=True,
    support_dof=False,
):
    depth_filenames = []
    if displacement_scale > 0:
        for i, slice_image in enumerate(state.image_slices):
            print(f"Generating depth map for slice {i}")
            depth_filename = state.depth_filename(i)
            if not depth_filename.exists():
                model = DepthEstimationModel(model=modelname)
                if model != state.depth_estimation_model:
                    state.depth_estimation_model = model
                depth_map = generate_depth_map(
                    slice_image.image[:, :, :3], model=state.depth_estimation_model
                )
                depth_map = postprocess_depth_map(
                    depth_map, slice_image.image[:, :, 3], final_blur=50
                )
                Image.fromarray(depth_map).save(depth_filename, compress_level=1)
            depth_filenames.append(depth_filename)

    # check whether we have upscaled slices we should use
    slices_filenames = []
    for i, slice_image in enumerate(state.image_slices):
        upscaled_filename = state.upscaled_filename(i)
        if upscaled_filename.exists():
            slices_filenames.append(upscaled_filename)
        else:
            slices_filenames.append(slice_image.filename)

    output_path = Path(filename) / state.MODEL_FILE
    gltf_path = export_gltf(
        output_path,
        camera,
        state.image_slices,
        slices_filenames,
        depth_filenames,
        displacement_scale=displacement_scale,
        inline_images=inline_images,
        support_dof=support_dof,
    )

    return gltf_path


@app.callback(
    Output(C.DOWNLOAD_IMAGE, "data"),
    Input({"type": "slice-info", "index": ALL}, "n_clicks"),
    State(C.STORE_APPSTATE_FILENAME, "data"),
    prevent_initial_call=True,
)
def download_image(n_clicks, filename):
    if filename is None or n_clicks is None or ctx.triggered_id is None:
        raise PreventUpdate()

    state = AppState.from_cache(filename)

    index = ctx.triggered_id["index"]
    if n_clicks[index] is None:
        raise PreventUpdate()

    # print(n_clicks, index, ctx.triggered)

    image_path = state.image_slices[index].filename

    return dcc.send_file(image_path, Path(state.image_slices[index].filename).name)


@app.callback(
    Output(C.STORE_UPDATE_SLICE, "data", allow_duplicate=True),
    Output(C.LOGS_DATA, "data", allow_duplicate=True),
    Input({"type": C.UPLOAD_SLICE, "index": ALL}, "contents"),
    State(C.STORE_APPSTATE_FILENAME, "data"),
    State(C.LOGS_DATA, "data"),
    prevent_initial_call=True,
)
def slice_upload(contents, filename, logs):
    if filename is None:
        raise PreventUpdate()

    state = AppState.from_cache(filename)
    if len(state.image_slices) == 0:
        raise PreventUpdate()

    index = ctx.triggered_id["index"]
    if contents[index] is None:
        raise PreventUpdate()

    # current aspect ratio for the given slice
    aspect_ratio = (
        state.image_slices[index].image.shape[1]
        / state.image_slices[index].image.shape[0]
    )

    content = contents[index]
    image = Image.open(io.BytesIO(base64.b64decode(content.split(",")[1])))
    image = image.convert("RGBA")

    if image.size[0] / image.size[1] != aspect_ratio:
        logs.append(
            "Fixing aspect ratio from {image.size[0] / image.size[1]} to {aspect_ratio}"
        )
        image = image.resize((int(aspect_ratio * image.size[1]), image.size[1]))

    image_filename = state.image_slices[index].new_version(np.array(image))
    state.to_file(
        filename, save_image_slices=False, save_depth_map=False, save_input_image=False
    )

    composed_image = state.image_slices[0].image.copy()
    for i, slice_image in enumerate(state.image_slices[1:]):
        blend_with_alpha(composed_image, slice_image.image)
    state.imgData = Image.fromarray(composed_image)

    logs.append(f"Received image slice upload for slice {index} at {image_filename}")

    return True, logs


@app.callback(
    Output(C.LOGS_DATA, "data", allow_duplicate=True),
    Output(C.ANIMATION_OUTPUT, "children"),
    Input(C.BTN_EXPORT_ANIMATION, "n_clicks"),
    State(C.STORE_APPSTATE_FILENAME, "data"),
    State(C.SLIDER_NUM_FRAMES, "value"),
    State(C.LOGS_DATA, "data"),
    running=[(Output(C.BTN_EXPORT_ANIMATION, "disabled"), True, False)],
    prevent_initial_call=True,
)
def export_animation(n_clicks, filename, num_frames, logs):
    if n_clicks is None or filename is None:
        raise PreventUpdate()

    state = AppState.from_cache(filename)

    camera_distance = state.camera.camera_distance

    camera_matrix = state.camera_matrix()
    card_corners_3d_list = state.get_cards()

    # Render the initial view
    camera_position = np.array([0, 0, -camera_distance], dtype=np.float32)
    render_image_sequence(
        filename,
        state.image_slices,
        card_corners_3d_list,
        camera_matrix,
        camera_position,
        push_distance=camera_distance * 0.75,  # XXX - make configurable
        num_frames=num_frames,
    )

    logs.append(f"Exported {num_frames} frames to animation")

    return logs, ""


@app.callback(
    Output(C.STORE_INPAINTING, "data", allow_duplicate=True),
    Input(C.STORE_RESTORE_STATE, "data"),
    prevent_initial_call=True,
)
def restore_inpainting(value):
    return True


@app.callback(
    Output(C.SLIDER_CAMERA_DISTANCE, "value"),
    Output(C.SLIDER_FOCAL_LENGTH, "value"),
    Output(C.SLIDER_MAX_DISTANCE, "value"),
    Output(C.SLIDER_DISPLACEMENT, "value"),
    Input(C.STORE_RESTORE_STATE, "data"),
    State(C.STORE_APPSTATE_FILENAME, "data"),
    prevent_initial_call=True,
)
def remember_camera_parameters(value, filename):
    if filename is None:
        raise PreventUpdate()

    state = AppState.from_cache(filename)
    cam = state.camera
    return (
        cam.camera_distance,
        cam.focal_length,
        cam.max_distance,
        state.mesh_displacement,
    )


@app.callback(
    Output(C.BTN_DARK_MODE, "n_clicks"),
    Input(C.STORE_RESTORE_STATE, "data"),
    State(C.STORE_APPSTATE_FILENAME, "data"),
    prevent_initial_call=True,
)
def restore_dark_mode(value, filename):
    if filename is None:
        raise PreventUpdate()

    state = AppState.from_cache(filename)
    return 1 if state.dark_mode else 0


@app.callback(
    Output(C.INPUT_API_KEY, "value"),
    Input(C.STORE_RESTORE_STATE, "data"),
    State(C.STORE_APPSTATE_FILENAME, "data"),
    prevent_initial_call=True,
)
def restore_api_key(value, filename):
    if filename is None:
        raise PreventUpdate()

    state = AppState.from_cache(filename)
    if state.api_key == None:
        return no_update

    return state.api_key


@app.callback(
    Output(C.UPLOAD_COMFYUI_WORKFLOW, "contents", allow_duplicate=True),
    Output(C.UPLOAD_COMFYUI_WORKFLOW, "filename", allow_duplicate=True),
    Input(C.STORE_RESTORE_STATE, "data"),
    State(C.STORE_APPSTATE_FILENAME, "data"),
    prevent_initial_call=True,
)
def restore_workflow(value, filename):
    if filename is None:
        raise PreventUpdate()

    state = AppState.from_cache(filename)
    if state.workflow_path().exists():
        workflow = state.workflow_path().read_bytes()
        return to_data_url(workflow), state.workflow_path().name
    return no_update, no_update


@app.callback(
    Output(C.DROPDOWN_DEPTH_MODEL, "value"),
    Output(C.DROPDOWN_INPAINT_MODEL, "value"),
    Input(C.STORE_RESTORE_STATE, "data"),
    State(C.STORE_APPSTATE_FILENAME, "data"),
    prevent_initial_call=True,
)
def restore_models(value, filename):
    if filename is None:
        raise PreventUpdate()

    state = AppState.from_cache(filename)

    depth_model_name = state.depth_model_name if state.depth_model_name else no_update
    inpainting_model_name = (
        state.inpainting_model_name if state.inpainting_model_name else no_update
    )

    return depth_model_name, inpainting_model_name


@app.callback(
    Output(C.INPUT_EXTERNAL_SERVER, "value"),
    Input(C.STORE_RESTORE_STATE, "data"),
    State(C.STORE_APPSTATE_FILENAME, "data"),
    prevent_initial_call=True,
)
def update_external_server_address(value, filename):
    if filename is None:
        raise PreventUpdate()

    state = AppState.from_cache(filename)
    return state.server_address


@app.callback(
    Output("model-viewer", "srcDoc"),
    Input(C.STORE_RESTORE_STATE, "data"),
    State(C.STORE_APPSTATE_FILENAME, "data"),
    prevent_initial_call=True,
)
def update_model_viewer(value, filename):
    if filename is None:
        raise PreventUpdate()

    state = AppState.from_cache(filename)

    iframe = get_gltf_iframe(state.serve_model_file())

    return iframe


@app.callback(
    Output(C.STORE_UPDATE_SLICE, "data", allow_duplicate=True),
    Input(C.STORE_RESTORE_STATE, "data"),
    State(C.STORE_APPSTATE_FILENAME, "data"),
    prevent_initial_call=True,
)
def restore_state_slices(value, filename):
    if filename is None:
        raise PreventUpdate()

    state = AppState.from_cache(filename)
    if len(state.image_slices) == 0:
        print("No image slices to restore")
        raise PreventUpdate()

    return True


@app.callback(
    Output(C.STORE_TRIGGER_UPDATE_DEPTHMAP, "data", allow_duplicate=True),
    Input(C.STORE_RESTORE_STATE, "data"),
    prevent_initial_call=True,
)
def restore_state_depthmap(value):
    return True


@app.callback(
    # XXX - generate depth-map via separate callback
    Output(C.STORE_APPSTATE_FILENAME, "data"),
    Output(C.STORE_RESTORE_STATE, "data"),
    Output(C.IMAGE, "src", allow_duplicate=True),
    Output(C.STORE_UPDATE_THRESHOLD_CONTAINER, "data", allow_duplicate=True),
    Output(C.SLIDER_NUM_SLICES, "value"),
    Output(C.LOGS_DATA, "data", allow_duplicate=True),
    Input(C.UPLOAD_STATE, "contents"),
    State(C.LOGS_DATA, "data"),
    prevent_initial_call=True,
)
def restore_state(contents, logs):
    if contents is None:
        raise PreventUpdate()

    # decode the contents into json
    content_type, content_string = contents.split(",")
    decoded_contents = base64.b64decode(content_string).decode("utf-8")
    state = AppState.from_json(decoded_contents)
    state.fill_from_files(state.filename)
    AppState.cache[state.filename] = state  # XXX - this may be too hacky

    logs.append(f"Restored state from {state.filename}")

    # XXX - refactor this to be triggered by a write to restore-state
    buffered = io.BytesIO()
    state.imgData.save(buffered, format="PNG")
    img_data = base64.b64encode(buffered.getvalue()).decode("utf-8")
    img_data = f"data:image/png;base64,{img_data}"

    return state.filename, True, img_data, True, state.num_slices, logs


@app.callback(
    Output(C.LOGS_DATA, "data", allow_duplicate=True),
    Input(C.BTN_SAVE_STATE, "n_clicks"),
    State(C.STORE_APPSTATE_FILENAME, "data"),
    State(C.LOGS_DATA, "data"),
    prevent_initial_call=True,
)
def save_state(n_clicks, filename, logs):
    if n_clicks is None or filename is None:
        raise PreventUpdate()

    state = AppState.from_cache(filename)
    state.to_file(filename)

    logs.append(f"Saved state to {filename}")

    return logs


@app.callback(
    Output(C.STORE_CURRENT_TAB, "data", allow_duplicate=True),
    Input({"type": "tab-content-main", "index": ALL}, "className"),
    prevent_initial_call=True,
)
def update_current_tab(classnames):
    names = ["Mode", "Segmentation", "Inpainting", "Export", "Configuration"]
    for i, classname in enumerate(classnames):
        if "hidden" not in classname:
            return names[i]
    return no_update


def main():
    """Main entry point for the Parallax Maker web application."""
    import argparse

    os.environ["DISABLE_TELEMETRY"] = "YES"
    os.environ["PYTORCH_ENABLE_MPS_FALLBACK"] = "1"

    # parse command line arguments
    parser = argparse.ArgumentParser(
        description="Parallax Maker - Turn images into 2.5D animations"
    )
    parser.add_argument(
        "--port", type=int, default=8050, help="Port to run the web server on"
    )
    parser.add_argument(
        "--host", type=str, default="127.0.0.1", help="Host to bind the web server to"
    )
    parser.add_argument(
        "--prefetch-models",
        type=str,
        default=None,
        help='Either "all" or "default" to prefetch models',
    )
    parser.add_argument("--debug", action="store_true", help="Run in debug mode")
    args = parser.parse_args()

    if not serving.is_running_from_reloader():
        if args.prefetch_models in ["all", "default"]:
            print("Prefetching models")
            if args.prefetch_models == "all":
                for model in [DepthEstimationModel, SegmentationModel, InpaintingModel]:
                    for model_name in model.MODELS:
                        model(model_name).load_model()
            else:
                DepthEstimationModel().load_model()
                SegmentationModel().load_model()
                InpaintingModel().load_model()
        elif args.prefetch_models is not None:
            print(
                f'Invalid prefetch models argument: {args.prefetch_models}; use "all" or "default"'
            )
            exit(1)

    print(f"Starting Parallax Maker on http://{args.host}:{args.port}")
    app.run_server(host=args.host, port=args.port, debug=args.debug)


if __name__ == "__main__":
    main()



"""Parallax Maker - A workflow for turning images into 2.5D animations.

This package provides tools for:
- Depth estimation from images using models like MiDaS and ZoeDepth
- Image segmentation using Segment Anything with point selection
- Inpainting with various models including Stable Diffusion
- 3D export to glTF format for Blender and Unreal Engine
- Web-based user interface for interactive workflow
"""

__version__ = "1.0.1"
__author__ = "Niels Provos"
__email__ = "niels@provos.org"
__license__ = "AGPL-3.0-or-later"

from .controller import AppState, CompositeMode
from .camera import Camera
from .slice import ImageSlice

__all__ = [
    "AppState",
    "CompositeMode",
    "Camera",
    "ImageSlice",
]



# (c) 2024 Niels Provos
#
"""
Provides instance segmentation functionality using pre-trained deep learning models.
Segment Anything seems superior to Mask2Former.

Supports two models:
- Mask2Former: https://huggingface.co/facebook/mask2former-swin-large-coco-instance
- SAM: https://huggingface.co/facebook/sam-vit-huge

Todo:
 - Create support for HQ-SAM: https://github.com/SysCV/sam-hq?tab=readme-ov-file
"""

from PIL import Image
import torch
from transformers import (
    AutoImageProcessor,
    Mask2FormerForUniversalSegmentation,
    SamModel,
    SamProcessor,
)
import numpy as np

from .utils import torch_get_device, image_overlay, draw_circle


class SegmentationModel:
    MODELS = ["mask2former", "sam"]

    def __init__(self, model="sam"):
        assert model in self.MODELS
        self.model_name = model
        self.model = None
        self.image_processor = None
        self.image = None
        self.mask = None

    def __eq__(self, other):
        if not isinstance(other, SegmentationModel):
            return False
        return self.model_name == other.model_name

    def load_model(self):
        load_model = {
            "mask2former": self.load_mask2former_model,
            "sam": self.load_sam_model,
        }

        result = load_model[self.model_name]()

        if self.model_name == "mask2former":
            self.model, self.image_processor = result
        elif self.model_name == "sam":
            self.model, self.image_processor = result

    @staticmethod
    def load_mask2former_model():
        image_processor = AutoImageProcessor.from_pretrained(
            "facebook/mask2former-swin-large-coco-instance"
        )
        model = Mask2FormerForUniversalSegmentation.from_pretrained(
            "facebook/mask2former-swin-large-coco-instance"
        )
        model.to(torch_get_device())
        return model, image_processor

    @staticmethod
    def load_sam_model():
        model = SamModel.from_pretrained("facebook/sam-vit-huge").to(torch_get_device())
        processor = SamProcessor.from_pretrained("facebook/sam-vit-huge")
        return model, processor

    def segment_image(self, image):
        self.image = image
        if isinstance(image, np.ndarray):
            self.image = Image.fromarray(image)
        self.image = self.image.convert("RGB")

        run_pipeline = {
            "mask2former": self.segment_image_mask2former,
            "sam": self.segment_image_sam,
        }

        self.mask = run_pipeline[self.model_name]()
        return self.mask

    def _get_mask_at_point_function(self):
        needs_mask = {"mask2former": True, "sam": False}
        if needs_mask[self.model_name] and self.mask is None:
            return None

        run_pipeline = {
            "mask2former": self.mask_at_point_mask2former,
            "sam": self.mask_at_point_sam,
        }
        return run_pipeline[self.model_name]

    def mask_at_point(self, point_xy):
        executor = self._get_mask_at_point_function()
        if executor is None:
            return None
        return executor(point_xy)

    def mask_at_point_blended(self, point_xy):
        executor = self._get_mask_at_point_function()

        # Apply segmentation function to each transformed image
        transforms = [
            "identity",
            "rotate_90",
            "rotate_180",
            "rotate_270",
            "flip_h",
            "flip_v",
        ]
        computed_masks = []
        image = self.image.copy()
        for transformation in transforms:
            transformed_image = SegmentationModel._transform_image(
                image, transformation
            )
            transformed_point = SegmentationModel._transform_point(
                point_xy, transformation, image.size
            )
            self.segment_image(transformed_image)
            mask = executor(transformed_point)
            if mask is not None:
                mask = SegmentationModel._inverse_transform(mask, transformation)
                mask_size = (mask.shape[1], mask.shape[0])
                assert (
                    mask_size == image.size
                ), f"Mask size {mask_size} does not match image size {image.size} for transformation {transformation}"
                computed_masks.append(mask)

        self.image = image

        computed_masks = SegmentationModel._filter_mask(computed_masks)
        blended_mask = np.array(computed_masks[0]).astype(np.float16)
        for mask in computed_masks[1:]:
            # print the max mask value in the console
            blended_mask += np.array(mask).astype(np.float16)

        blended_mask /= len(computed_masks)
        blended_mask = blended_mask.astype(np.uint8)

        return blended_mask

    def segment_image_mask2former(self):
        if self.model is None:
            self.load_model()

        inputs = self.image_processor(self.image, size=1280, return_tensors="pt").to(
            self.model.device
        )
        with torch.no_grad():
            outputs = self.model(**inputs)

        pred_map = self.image_processor.post_process_instance_segmentation(
            outputs, target_sizes=[self.image.size[::-1]]
        )[0]
        self.mask = pred_map["segmentation"]
        return self.mask

    def mask_at_point_mask2former(self, point_xy):
        width, height = self.image.size
        assert point_xy[0] < width and point_xy[1] < height

        label = int(self.mask[point_xy[1], point_xy[0]])
        if label == -1:
            return None

        mask = (self.mask == label).numpy().astype(np.uint8)
        mask *= 255

        return mask

    def segment_image_sam(self):
        # this is a no-op for sam as the model needs to be guided by an input point
        return None

    def mask_at_point_sam(self, point_input):
        """
        Masks the given points using the SAM algorithm.

        Args:
            point_input (tuple, list, dict): The input points to be masked. It can be one of the following:
                - tuple: A single point.
                - list of tuples: Multiple points.
                - dict with 'positive_points' and 'negative_points' keys: Lists of positive and negative points.

        Returns:
            list: The masked points.

        Raises:
            ValueError: If the input is invalid.

        """
        positive_points = []
        negative_points = []
        if isinstance(point_input, tuple):
            # Single point case
            positive_points = [point_input]
        elif isinstance(point_input, list) and all(
            isinstance(item, tuple) for item in point_input
        ):
            # List of points case
            positive_points = point_input
        elif (
            isinstance(point_input, dict)
            and "positive_points" in point_input
            and "negative_points" in point_input
        ):
            # Lists of positive and negative points case
            positive_points = point_input["positive_points"]
            negative_points = point_input["negative_points"]
        else:
            raise ValueError("Invalid input for mask_at_point function.")
        return self._mask_at_point_sam(positive_points, negative_points)

    def _mask_at_point_sam(self, positive_points, negative_points):
        if self.model is None:
            self.load_model()

        input_points = positive_points + negative_points
        input_labels = [1] * len(positive_points) + [0] * len(negative_points)
        inputs = self.image_processor(
            self.image,
            input_points=[input_points],
            input_labels=[input_labels],
            return_tensors="pt",
        )
        # convert inputs to dtype torch.float32
        inputs = inputs.to(torch.float32).to(self.model.device)

        with torch.no_grad():
            outputs = self.model(**inputs)

        # the model is capabale of returning multiple masks, but we only return one for now
        masks = self.image_processor.image_processor.post_process_masks(
            outputs.pred_masks.cpu(),
            inputs["original_sizes"].cpu(),
            inputs["reshaped_input_sizes"].cpu(),
        )
        scores = outputs.iou_scores.cpu()

        mask = masks[0].numpy()

        # format the mask for visualization
        if len(mask.shape) == 4:
            mask = mask.squeeze()
        if scores.shape[0] == 1:
            scores = scores.squeeze()

        # 3 predictions at this point?
        # find the index where scores has the highest value
        index = int(torch.argmax(scores))
        mask = mask[index]

        h, w = mask.shape[-2:]
        mask_image = mask.reshape(h, w)
        mask_image = (mask_image * 255).astype(np.uint8)

        return mask_image

    # Function to rotate point around the image center for specific angles
    @staticmethod
    def _rotate_point(point, angle, image_size):
        assert angle in [0, 90, 180, 270]
        ox, oy = image_size[0] // 2, image_size[1] // 2
        x, y = point[0] - ox, point[1] - oy

        if angle == 90:
            new_x, new_y = oy - y, x + ox
        elif angle == 180:
            new_x, new_y = ox - x, oy - y
        elif angle == 270:
            new_x, new_y = y + oy, ox - x
        else:
            new_x, new_y = point

        return int(new_x), int(new_y)

    @staticmethod
    def _transform_image(image, transformation):
        if transformation == "rotate_90":
            return image.rotate(-90, expand=True)
        elif transformation == "rotate_180":
            return image.rotate(-180, expand=True)
        elif transformation == "rotate_270":
            return image.rotate(-270, expand=True)
        elif transformation == "flip_h":
            return image.transpose(Image.FLIP_LEFT_RIGHT)
        elif transformation == "flip_v":
            return image.transpose(Image.FLIP_TOP_BOTTOM)
        else:
            return image

    @staticmethod
    def _transform_point(point, transformation, image_size):
        if isinstance(point, tuple):
            return SegmentationModel._transform_point_single(
                point, transformation, image_size
            )
        elif isinstance(point, list) and all(isinstance(item, tuple) for item in point):
            transformed_points = []
            for p in point:
                transformed_points.append(
                    SegmentationModel._transform_point_single(
                        p, transformation, image_size
                    )
                )
            return transformed_points
        elif (
            isinstance(point, dict)
            and "positive_points" in point
            and "negative_points" in point
        ):
            positive_points = point["positive_points"]
            negative_points = point["negative_points"]
            transformed_points = {"positive_points": [], "negative_points": []}
            for p in positive_points:
                transformed_points["positive_points"].append(
                    SegmentationModel._transform_point_single(
                        p, transformation, image_size
                    )
                )
            for n in negative_points:
                transformed_points["negative_points"].append(
                    SegmentationModel._transform_point_single(
                        n, transformation, image_size
                    )
                )
            return transformed_points
        else:
            raise ValueError("Invalid input for _transform_point function.")

    @staticmethod
    def _transform_point_single(point, transformation, image_size):
        if transformation == "rotate_90":
            return SegmentationModel._rotate_point(point, 90, image_size)
        elif transformation == "rotate_180":
            return SegmentationModel._rotate_point(point, 180, image_size)
        elif transformation == "rotate_270":
            return SegmentationModel._rotate_point(point, 270, image_size)
        elif transformation == "flip_h":
            return (image_size[0] - point[0], point[1])
        elif transformation == "flip_v":
            return (point[0], image_size[1] - point[1])
        else:
            return point

    @staticmethod
    def _inverse_transform(mask, transformation):
        mask = Image.fromarray(mask)
        if transformation == "rotate_90":
            result = mask.rotate(90, expand=True)
        elif transformation == "rotate_180":
            result = mask.rotate(180, expand=True)
        elif transformation == "rotate_270":
            result = mask.rotate(270, expand=True)
        elif transformation == "flip_h":
            result = mask.transpose(Image.FLIP_LEFT_RIGHT)
        elif transformation == "flip_v":
            result = mask.transpose(Image.FLIP_TOP_BOTTOM)
        else:
            result = mask

        result = np.array(result)

        return result

    @staticmethod
    def _filter_mask(masks):
        selected_pixels = []
        for mask in masks:
            # compute the number of pixels in the mask
            num_pixels = np.sum(mask)
            selected_pixels.append(num_pixels)

        # compute the median and standard deviation of the number of pixels
        median_pixels = np.median(selected_pixels)
        std_pixels = np.std(selected_pixels)

        print(f"Median pixels: {median_pixels}, Std pixels: {std_pixels}")

        # filter out masks that have less than the median number of pixels
        filtered_masks = []
        for mask in masks:
            num_pixels = np.sum(mask)
            if (
                num_pixels >= median_pixels - std_pixels
                and num_pixels <= median_pixels + std_pixels
            ):
                filtered_masks.append(mask)
            else:
                print(f"Filtering mask with {num_pixels} pixels")

        return filtered_masks


if __name__ == "__main__":
    filename = "input.jpg"

    model = SegmentationModel()
    image = Image.open(filename)

    # Apply segmentation function to original image
    point_xy = (500, 600)
    negative_point_xy = (200, 600)
    mask = model.segment_image(image)

    point_input = {
        "positive_points": [point_xy],
        "negative_points": [negative_point_xy],
    }

    mask_at_point = model.mask_at_point_blended(point_input)
    mask = Image.fromarray(mask_at_point).convert("RGB")

    result = image_overlay(image, mask)

    # draw a circle at the point on image
    draw_circle(result, point_xy, 20, fill_color=(0, 255, 0))
    draw_circle(result, negative_point_xy, 20, fill_color=(255, 0, 0))

    result.save("segmented_image.png")



import numpy as np
import unittest

from .inpainting import find_nearest_alpha


class TestFindNearestAlpha(unittest.TestCase):
    def test_nearest_alpha_with_full_alpha(self):
        alpha = np.ones((3, 3)) * 255
        nearest_alpha = np.full((3, 3, 4, 2), -1, dtype=int)
        expected = np.zeros((3, 3, 4, 2))
        for i in range(3):
            for j in range(3):
                expected[i, j, :, :] = [[i, j], [i, j], [i, j], [i, j]]
        find_nearest_alpha(alpha, nearest_alpha)
        np.testing.assert_array_equal(nearest_alpha, expected)

    def test_nearest_alpha_with_partial_alpha(self):
        alpha = np.array([[255, 0, 255], [255, 255, 0], [0, 255, 255]])
        nearest_alpha = np.full((3, 3, 4, 2), -1, dtype=int)
        expected = np.zeros((3, 3, 4, 2))
        expected[0, 0, :, :] = [[0, 0], [0, 0], [0, 0], [0, 0]]
        expected[0, 1, :, :] = [[-1, -1], [0, 0], [1, 1], [0, 2]]
        expected[0, 2, :, :] = [[0, 2], [0, 2], [0, 2], [0, 2]]
        expected[1, 0, :, :] = [[1, 0], [1, 0], [1, 0], [1, 0]]
        expected[1, 1, :, :] = [[1, 1], [1, 1], [1, 1], [1, 1]]
        expected[1, 2, :, :] = [[0, 2], [1, 1], [2, 2], [-1, -1]]
        expected[2, 0, :, :] = [[1, 0], [-1, -1], [-1, -1], [2, 1]]
        expected[2, 1, :, :] = [[2, 1], [2, 1], [2, 1], [2, 1]]
        expected[2, 2, :, :] = [[2, 2], [2, 2], [2, 2], [2, 2]]
        find_nearest_alpha(alpha, nearest_alpha)
        np.testing.assert_array_equal(nearest_alpha, expected)


if __name__ == "__main__":
    unittest.main()



import numpy as np
import unittest
from unittest.mock import mock_open, patch, MagicMock
from pathlib import Path
from PIL import Image
import numpy as np

from .utils import encode_string_with_nonce, decode_string_with_nonce
from .controller import AppState
from .slice import ImageSlice


class TestAddSlice(unittest.TestCase):
    def setUp(self) -> None:
        super().setUp()

        self.state = AppState()
        self.state.filename = "test"

    def test_add_slice_normal(self):
        # Create a mock slice image and depth value
        slice_image1 = np.ones((10, 10, 4), np.uint8)
        depth1 = 5

        image_slice = ImageSlice(slice_image1, depth1)

        # Call the function
        insert1 = self.state.add_slice(image_slice)

        self.assertEqual(insert1, 0)
        self.assertEqual(len(self.state.image_slices), 1)
        self.assertEqual(self.state.image_slices[0].depth, depth1)

        slice_image2 = np.ones((10, 10, 4), np.uint8)
        depth2 = 10
        image_slice = ImageSlice(slice_image2, depth2)

        # Call the function
        insert2 = self.state.add_slice(image_slice)

        # Check whether this slice was added before the first slice
        self.assertEqual(insert2, 1)
        self.assertEqual(len(self.state.image_slices), 2)
        self.assertEqual(self.state.image_slices[0].depth, depth1)
        self.assertEqual(self.state.image_slices[1].depth, depth2)

    def test_add_slice_reverse(self):
        # Create a mock slice image and depth value
        slice_image1 = np.ones((10, 10, 4), np.uint8)
        depth1 = 5

        image_slice1 = ImageSlice(slice_image1, depth1)

        # Call the function
        insert1 = self.state.add_slice(image_slice1)

        self.assertEqual(insert1, 0)
        self.assertEqual(len(self.state.image_slices), 1)
        self.assertEqual(self.state.image_slices[0].depth, depth1)

        slice_image2 = np.ones((10, 10, 4), np.uint8)
        depth2 = 1

        image_slice2 = ImageSlice(slice_image2, depth2)

        # Call the function
        insert2 = self.state.add_slice(image_slice2)

        # Check whether this slice was added before the first slice
        self.assertEqual(insert2, 0)
        self.assertEqual(len(self.state.image_slices), 2)
        self.assertEqual(self.state.image_slices[0].depth, depth2)
        self.assertEqual(self.state.image_slices[1].depth, depth1)

        slice_image3 = np.ones((10, 10, 4), np.uint8)
        depth3 = 1

        image_slice3 = ImageSlice(slice_image3, depth3)

        # Call the function
        insert3 = self.state.add_slice(image_slice3)

        # Check that the slice was added and depth increased to avoid duplicates
        self.assertEqual(insert3, 1)
        self.assertEqual(len(self.state.image_slices), 3)
        self.assertEqual(self.state.image_slices[0].depth, depth2)
        self.assertEqual(self.state.image_slices[1].depth, depth3 + 1)

        # check that all the filenames are accurate
        expected = [
            "test/image_slice_1.png",
            "test/image_slice_2.png",
            "test/image_slice_0.png",
        ]

        image_slices_filenames = [slice.filename for slice in self.state.image_slices]

        self.assertEqual(image_slices_filenames, expected)


class TestCheckPathnames(unittest.TestCase):
    def setUp(self) -> None:
        super().setUp()

        self.state = AppState()
        self.state.filename = "appstate-test"

    def test_check_pathnames_valid(self):
        # Set up valid pathnames
        image_slices_filenames = [
            "appstate-test/image_slice_1.png",
            "appstate-test/image_slice_2.png",
            "appstate-test/image_slice_0.png",
        ]

        self.state.image_slices = [
            ImageSlice(filename=filename) for filename in image_slices_filenames
        ]

        # Call the function
        self.state.check_pathnames()

    def test_check_pathnames_invalid_root(self):
        # Set up invalid root pathname
        self.state.filename = "invalid_root/../something"
        image_slices_filenames = [
            "appstate-test/image_slice_2.png",
            "appstate-test/image_slice_0.png",
        ]

        self.state.image_slices = [
            ImageSlice(filename=filename) for filename in image_slices_filenames
        ]

        # Call the function and expect an AssertionError
        with self.assertRaises(AssertionError):
            self.state.check_pathnames()

    def test_check_pathnames_invalid_filenames(self):
        # Set up invalid filenames
        image_slices_filenames = [
            "appstate-test/image_slice_1.png",
            "appstate-test/image_slice_2.png",
            "appstate-test/../../../invalid_filename.png",
        ]

        self.state.image_slices = [
            ImageSlice(filename=filename) for filename in image_slices_filenames
        ]

        # Call the function and expect an AssertionError
        with self.assertRaises(AssertionError):
            self.state.check_pathnames()


class TestChangeSliceDepth(unittest.TestCase):
    def setUp(self) -> None:
        super().setUp()

        self.state = AppState()
        self.state.filename = "test"

        # Add some initial slices
        self.state.add_slice(
            ImageSlice(
                np.ones((10, 10, 4), np.uint8),
                5,
                positive_prompt="one",
                negative_prompt="two",
            )
        )
        self.state.add_slice(
            ImageSlice(
                np.ones((10, 10, 4), np.uint8),
                10,
                positive_prompt="three",
                negative_prompt="four",
            )
        )
        self.state.add_slice(
            ImageSlice(
                np.ones((10, 10, 4), np.uint8),
                15,
                positive_prompt="blue",
                negative_prompt="green",
            )
        )

    def test_change_slice_depth_same_depth(self):
        # Get the initial state
        initial_depths = [image_slice.depth for image_slice in self.state.image_slices]

        # Change the depth of the second slice to the same depth
        slice_index = 1
        depth = 10
        new_index = self.state.change_slice_depth(slice_index, depth)

        # Check that the slice index remains the same
        self.assertEqual(new_index, slice_index)

        # Check that the slices and depths remain unchanged
        new_depths = [image_slice.depth for image_slice in self.state.image_slices]
        self.assertEqual(new_depths, initial_depths)

    def test_change_slice_depth_valid(self):
        # Get the initial state
        initial_slices = self.state.image_slices.copy()

        # Change the depth of the second slice to a new depth
        slice_index = 1
        depth = 3
        new_index = self.state.change_slice_depth(slice_index, depth)

        # Check that the new index is returned
        self.assertEqual(new_index, 0)

        # Check that the depth of the second slice is updated
        image_slice = self.state.image_slices[new_index]
        self.assertEqual(image_slice.depth, depth)
        self.assertEqual(image_slice.positive_prompt, "three")
        self.assertEqual(image_slice.negative_prompt, "four")

        # Check that the other slices and depths remain unchanged
        self.assertEqual(self.state.image_slices[1], initial_slices[0])
        self.assertEqual(self.state.image_slices[2], initial_slices[2])

    def test_change_slice_depth_invalid_index(self):
        # Get the initial state
        initial_slices = self.state.image_slices.copy()

        # Try to change the depth of an invalid slice index
        slice_index = 3
        depth = 12
        with self.assertRaises(AssertionError):
            self.state.change_slice_depth(slice_index, depth)

        # Check that the slices and depths remain unchanged
        self.assertEqual(self.state.image_slices, initial_slices)


class TestFromJson(unittest.TestCase):
    def test_from_json_empty(self):
        json_data = None
        state = AppState.from_json(json_data)
        self.assertIsInstance(state, AppState)
        self.assertEqual(state.image_slices, [])
        self.assertEqual(state.filename, None)
        self.assertEqual(state.num_slices, 5)
        self.assertEqual(state.imgThresholds, None)
        self.assertEqual(state.depth_model_name, None)
        self.assertEqual(state.inpainting_model_name, None)
        self.assertEqual(state.server_address, None)
        self.assertEqual(state.api_key, None)

    def test_from_json_full(self):
        encoded = encode_string_with_nonce("sk-somesecretkey", "appstate-test")
        json_data = """
        {
            "filename": "appstate-test",
            "num_slices": 10,
            "imgThresholds": [0, 255],
            "image_depths": [0, 1, 2],
            "image_slices_filenames": [
                "appstate-test/image1.png",
                "appstate-test/image2.png",
                "appstate-test/image3.png"],
            "depth_model_name": "model",
            "inpainting_model_name": "inpainting",
            "positive_prompts": ["one fish", "two fish", "red fish"],
            "negative_prompts": ["blue fish", "new fish", "old fish"],
            "server_address": "http://localhost:8000",
            "api_key": "ENCODED",
            "camera_distance": 1.0,
            "max_distance": 2.0,
            "focal_length": 3.0,
            "mesh_displacement": 4.0
        }
        """
        json_data = json_data.replace("ENCODED", encoded)

        state = AppState.from_json(json_data)
        self.assertIsInstance(state, AppState)
        self.assertEqual(state.filename, "appstate-test")
        self.assertEqual(state.num_slices, 10)
        self.assertEqual(state.imgThresholds, [0, 255])

        image_depths = [image_slice.depth for image_slice in state.image_slices]
        self.assertEqual(image_depths, [0, 1, 2])

        image_slices_filenames = [
            image_slice.filename for image_slice in state.image_slices
        ]
        self.assertEqual(
            image_slices_filenames,
            [
                "appstate-test/image1.png",
                "appstate-test/image2.png",
                "appstate-test/image3.png",
            ],
        )
        self.assertEqual(state.depth_model_name, "model")
        self.assertEqual(state.inpainting_model_name, "inpainting")

        positive_prompts = [
            image_slice.positive_prompt for image_slice in state.image_slices
        ]
        self.assertEqual(positive_prompts, ["one fish", "two fish", "red fish"])

        negative_prompts = [
            image_slice.negative_prompt for image_slice in state.image_slices
        ]
        self.assertEqual(negative_prompts, ["blue fish", "new fish", "old fish"])

        self.assertEqual(state.server_address, "http://localhost:8000")
        self.assertEqual(state.api_key, "sk-somesecretkey")
        self.assertEqual(state.camera.camera_distance, 1.0)
        self.assertEqual(state.camera.max_distance, 2.0)
        self.assertEqual(state.camera.focal_length, 3.0)
        self.assertEqual(state.mesh_displacement, 4.0)

    def test_from_json_partial(self):
        json_data = """
            {
                "filename": "appstate-test",
                "num_slices": 10,
                "imgThresholds": [0, 255],
                "image_depths": [0, 1, 2],
                "image_slices_filenames": [
                    "appstate-test/image1.png",
                    "appstate-test/image2.png",
                    "appstate-test/image3.png"]
            }
            """
        state = AppState.from_json(json_data)
        self.assertIsInstance(state, AppState)
        self.assertEqual(state.filename, "appstate-test")
        self.assertEqual(state.num_slices, 10)
        self.assertEqual(state.imgThresholds, [0, 255])

        image_depths = [image_slice.depth for image_slice in state.image_slices]
        self.assertEqual(image_depths, [0, 1, 2])

        image_slices_filenames = [
            image_slice.filename for image_slice in state.image_slices
        ]
        self.assertEqual(
            image_slices_filenames,
            [
                "appstate-test/image1.png",
                "appstate-test/image2.png",
                "appstate-test/image3.png",
            ],
        )
        self.assertEqual(state.depth_model_name, None)
        self.assertEqual(state.inpainting_model_name, None)

        positive_prompts = [
            image_slice.positive_prompt for image_slice in state.image_slices
        ]
        self.assertEqual(positive_prompts, [""] * 3)

        negative_prompts = [
            image_slice.negative_prompt for image_slice in state.image_slices
        ]
        self.assertEqual(negative_prompts, [""] * 3)
        self.assertEqual(state.server_address, None)
        self.assertEqual(state.api_key, None)


class TestUpscaling(unittest.TestCase):
    def setUp(self):
        self.state = AppState()
        self.image = np.zeros((100, 100, 3), dtype=np.uint8)

    def test_create_upscaler_default(self):
        self.state.pipeline_spec = None
        self.state._create_upscaler()
        self.assertIsNotNone(self.state.upscaler)

    @patch("parallax_maker.controller.StabilityAI")
    def test_create_upscaler_stabilityai(self, mock_stabilityai):
        self.state.pipeline_spec = MagicMock()
        self.state.inpainting_model_name = "stabilityai"
        self.state.api_key = "test_api_key"
        self.state._create_upscaler()
        self.assertIsNotNone(self.state.upscaler)
        mock_stabilityai.assert_called_with("test_api_key")

    def test_create_upscaler_inpainting(self):
        self.state.pipeline_spec = MagicMock()
        self.state.inpainting_model_name = "inpainting"
        self.state._create_upscaler()
        self.assertIsNotNone(self.state.upscaler)
        self.state.pipeline_spec.load_model.assert_called_once()

    @patch("parallax_maker.controller.Upscaler")
    def test_upscale_image(self, mock_upscaler):
        mock_upscaler_instance = MagicMock()
        mock_upscaler.return_value = mock_upscaler_instance
        upscaled_image = self.state.upscale_image(self.image)
        mock_upscaler_instance.upscale_image_tiled.assert_called_with(
            self.image, overlap=64, prompt="", negative_prompt=""
        )
        self.assertIsNotNone(upscaled_image)

    @patch("parallax_maker.controller.Upscaler")
    def test_upscale_image_with_prompts(self, mock_upscaler):
        mock_upscaler_instance = MagicMock()
        mock_upscaler.return_value = mock_upscaler_instance
        prompt = "test prompt"
        negative_prompt = "test negative prompt"
        self.state.upscale_image(self.image, prompt, negative_prompt)
        mock_upscaler_instance.upscale_image_tiled.assert_called_with(
            self.image, overlap=64, prompt=prompt, negative_prompt=negative_prompt
        )


class TestSaveImageSlices(unittest.TestCase):
    def setUp(self) -> None:
        super().setUp()

        self.state = AppState()
        self.state.filename = "appstate-random"

        # Add some initial slices
        self.state.add_slice(
            ImageSlice(
                np.zeros((10, 10, 4), np.uint8),
                5,
                positive_prompt="one",
                negative_prompt="two",
            )
        )
        self.state.add_slice(
            ImageSlice(
                np.zeros((10, 10, 4), np.uint8),
                10,
                positive_prompt="three",
                negative_prompt="four",
            )
        )
        self.state.add_slice(
            ImageSlice(
                np.zeros((10, 10, 4), np.uint8),
                15,
                positive_prompt="blue",
                negative_prompt="green",
            )
        )

    def test_save_image_slices(self):
        # Mock the image saving function
        with patch("parallax_maker.slice.Image.Image.save") as mock_imwrite:
            self.state.save_image_slices(self.state.filename)

            # Check that the image slices were saved
            self.assertEqual(mock_imwrite.call_count, 3)
            mock_imwrite.assert_any_call("appstate-random/image_slice_0.png")
            mock_imwrite.assert_any_call("appstate-random/image_slice_1.png")
            mock_imwrite.assert_any_call("appstate-random/image_slice_2.png")


class TestToFile(unittest.TestCase):

    def setUp(self):
        # Creating a mock AppState instance
        self.app_state = AppState()

        self.file_path = Path("appstate-random")

        # Mocking methods directly on the instance
        self.img_data_mock = MagicMock()
        self.app_state.imgData = self.img_data_mock

        self.depth_map_data_mock = MagicMock()
        self.app_state.depthMapData = self.depth_map_data_mock

    @patch("parallax_maker.controller.Image.fromarray")
    @patch("parallax_maker.controller.Path.mkdir")
    @patch("parallax_maker.controller.Path.exists")
    @patch("parallax_maker.controller.shutil.move")
    @patch("parallax_maker.controller.open", new_callable=mock_open)
    @patch("parallax_maker.controller.Path.unlink")
    @patch.object(AppState, "to_json", return_value='{"state": "dummy state"}')
    def test_to_file(
        self,
        mock_to_json,
        mock_unlink,
        mock_open_func,
        mock_shutil_move,
        mock_exists,
        mock_mkdir,
        mock_fromarray,
    ):
        mock_mkdir.return_value = None
        # file_path, state_file, backup_file exists
        mock_exists.side_effect = [True, True, True]
        mock_fromarray.return_value = MagicMock()

        self.app_state.to_file(self.file_path)

        # Check if the directory was created
        mock_mkdir.assert_called_once_with()

        # Check if files were saved
        self.img_data_mock.save.assert_called_once_with(
            str(self.file_path / AppState.IMAGE_FILE)
        )
        mock_fromarray.return_value.save.assert_called_once_with(
            str(self.file_path / AppState.DEPTH_MAP_FILE)
        )

        # Check if state file operations were performed correctly
        state_file = self.file_path / AppState.STATE_FILE
        temp_file = state_file.with_suffix(".tmp")
        backup_file = state_file.with_suffix(".bak")
        mock_open_func.assert_called_once_with(temp_file, "w", encoding="utf-8")

        # Read the written content
        handle = mock_open_func()
        handle.write.assert_called_once_with('{"state": "dummy state"}')

        mock_shutil_move.assert_any_call(state_file, backup_file)
        mock_shutil_move.assert_any_call(temp_file, state_file)
        mock_unlink.assert_called_once()

    @patch("parallax_maker.controller.Image.fromarray")
    @patch("parallax_maker.controller.Path.mkdir")
    @patch("parallax_maker.controller.Path.exists")
    @patch("parallax_maker.controller.shutil.move")
    @patch("parallax_maker.controller.open", new_callable=mock_open)
    @patch("parallax_maker.controller.Path.unlink")
    @patch.object(AppState, "to_json", return_value='{"state": "dummy state"}')
    def test_to_file_restore_backup_on_error(
        self,
        mock_to_json,
        mock_unlink,
        mock_open_func,
        mock_shutil_move,
        mock_exists,
        mock_mkdir,
        mock_fromarray,
    ):
        mock_mkdir.return_value = None
        # state_file, no backup file
        mock_exists.side_effect = [True, True, False]
        # First call succeeds, second (inside state write) fails
        mock_fromarray.return_value = MagicMock()

        # Configure mock_open to work with 'with' statement
        mock_file_handle = mock_open_func.return_value
        mock_file_handle.__enter__.return_value = mock_file_handle
        mock_open_func.side_effect = [IOError]

        with self.assertRaises(IOError):
            self.app_state.to_file(self.file_path)

        # Check if appropriate cleanup was done
        temp_file = (self.file_path / AppState.STATE_FILE).with_suffix(".tmp")
        backup_file = (self.file_path / AppState.STATE_FILE).with_suffix(".bak")

        mock_unlink.assert_called_once()
        mock_shutil_move.assert_called_once_with(
            backup_file, self.file_path / AppState.STATE_FILE
        )


class TestApplyMask(unittest.TestCase):
    def setUp(self):
        self.state = AppState()
        self.state.imgData = Image.new("RGB", (100, 100))
        self.state.result_tinted = Image.new("RGB", (100, 100), (255, 0, 0))
        self.state.grayscale_tinted = Image.new("RGB", (100, 100), (128, 128, 128))

    def test_apply_mask_with_pil_image(self):
        mask = Image.new("L", (100, 100), 128)

        result = self.state.apply_mask(self.state.imgData, mask)

        self.assertIsInstance(result, Image.Image)
        self.assertEqual(result.size, self.state.imgData.size)

    def test_apply_mask_with_numpy_array(self):
        mask = np.zeros((100, 100), dtype=np.uint8)

        result = self.state.apply_mask(self.state.imgData, mask)

        self.assertIsInstance(result, Image.Image)
        self.assertEqual(result.size, self.state.imgData.size)

    def test_apply_mask_with_none_image(self):
        image = None
        mask = Image.new("L", (100, 100))

        with self.assertRaises(AttributeError):
            self.state.apply_mask(image, mask)

    def test_apply_mask_creates_tints_when_needed(self):
        mask = Image.new("L", (100, 100), 128)
        self.state.result_tinted = None
        self.state.grayscale_tinted = None

        result = self.state.apply_mask(self.state.imgData, mask)

        self.assertIsInstance(result, Image.Image)
        self.assertEqual(result.size, self.state.imgData.size)

    def test_apply_mask_uses_precomputed_tints(self):
        mask = Image.new("L", (100, 100), 128)

        result = self.state.apply_mask(self.state.imgData, mask)

        self.assertNotEqual(self.state.result_tinted, None)
        self.assertNotEqual(self.state.grayscale_tinted, None)
        self.assertIsInstance(result, Image.Image)
        self.assertEqual(result.size, self.state.imgData.size)


if __name__ == "__main__":
    unittest.main()



import unittest
from .segmentation import (
    analyze_depth_histogram,
    blend_with_alpha,
    generate_simple_thresholds,
)
import numpy as np


class TestAnalyzeDepthHistogram(unittest.TestCase):
    def test_analyze_depth_histogram(self):
        # Create a dummy depth map
        depth_map = np.array([[0, 20, 30], [40, 50, 60], [70, 80, 255]])

        # Define the number of slices
        num_slices = 5

        # Call the function
        result = analyze_depth_histogram(depth_map, num_slices=num_slices)

        # Define the expected result
        expected_result = [0, 30, 40, 60, 70, 255]

        # Assert that the result is as expected
        self.assertEqual(result, expected_result)

    def test_analyze_depth_histogram_with_different_slices(self):
        # Create a dummy depth map
        depth_map = np.array([[0, 20, 30], [40, 50, 60], [70, 80, 255]])

        # Define the number of slices
        num_slices = 3

        # Call the function
        result = analyze_depth_histogram(depth_map, num_slices=num_slices)

        # Define the expected result
        expected_result = [0, 40, 70, 255]

        # Assert that the result is as expected
        self.assertEqual(result, expected_result)

    def test_analyze_depth_histogram_with_one_slice(self):
        # Create a dummy depth map
        depth_map = np.array([[0, 20, 30], [40, 50, 60], [70, 80, 255]])

        # Define the number of slices
        num_slices = 1

        # Call the function
        result = analyze_depth_histogram(depth_map, num_slices=num_slices)

        # Define the expected result
        expected_result = [0, 255]

        # Assert that the result is as expected
        self.assertEqual(result, expected_result)


class TestGenerateSimpleThresholds(unittest.TestCase):
    def test_generate_simple_thresholds(self):
        # Create a dummy image
        image = np.array([[0, 20, 30], [40, 50, 60], [70, 80, 255]])

        # Define the number of slices
        num_slices = 5

        # Call the function
        result = generate_simple_thresholds(image, num_slices=num_slices)

        # Define the expected result
        expected_result = np.array([0, 51, 102, 153, 204, 255])

        # Assert that the result is as expected
        np.testing.assert_array_equal(result, expected_result)


class TestBlendWithAlpha(unittest.TestCase):
    def test_blend_with_alpha(self):
        # Create target image
        target_image = np.array(
            [
                [[255, 0, 0, 255], [0, 255, 0, 255], [0, 0, 255, 255]],
                [[255, 255, 0, 255], [255, 0, 255, 255], [0, 255, 255, 255]],
                [[255, 255, 255, 255], [0, 0, 0, 255], [128, 128, 128, 255]],
            ]
        )

        # Create merge image
        merge_image = np.array(
            [
                [[0, 0, 0, 128], [128, 128, 128, 128], [255, 255, 255, 128]],
                [[128, 0, 0, 128], [0, 128, 0, 128], [0, 0, 128, 128]],
                [[128, 128, 0, 128], [128, 0, 128, 128], [0, 128, 128, 128]],
            ]
        )

        # Call the function
        blend_with_alpha(target_image, merge_image)

        # Define the expected result
        expected_result = np.array(
            [
                [[127, 0, 0, 255], [64, 191, 64, 255], [128, 128, 255, 255]],
                [[191, 127, 0, 255], [127, 64, 127, 255], [0, 127, 191, 255]],
                [[191, 191, 127, 255], [64, 0, 64, 255], [63, 128, 128, 255]],
            ]
        )

        # Assert that the result is as expected
        np.testing.assert_array_equal(target_image, expected_result)


if __name__ == "__main__":
    unittest.main()



LOGS_DATA = "logs-data"

ICON_DARK_MODE = "dark-mode-icon"
BTN_DARK_MODE = "dark-mode-toggle"

LOADING_UPLOAD = "loading-upload"
LOADING_DEPTHMAP = "loading-depthmap"
LOADING_GENERATE_SLICE = "loading-generate-slice"
LOADING_GENERATE_INPAINTING = "generate-inpainting"
LOADING_GLTF = "gltf-loading"
LOADING_ANIMATION = "animation-loading"

CTR_DEPTH_MAP = "depth-map-container"
BTN_GENERATE_DEPTHMAP = "generate-depthmap-button"
DEPTHMAP_OUTPUT = "gen-depthmap-output"
ANIMATION_OUTPUT = "gen-animation-output"

CANVAS_DATA = "canvas-data"
CANVAS_MASK_DATA = "canvas-mask-data"
CANVAS_PAINT = "canvas-paint"

STORE_SELECTED_SLICE = "selected-slice"
STORE_IGNORE = "store-ignore"
STORE_GENERATE_SLICE = "generate-slice-request"
STORE_UPDATE_SLICE = "update-slice-request"
STORE_INPAINTING = "inpainting-request"
STORE_APPSTATE_FILENAME = "application-state-filename"
STORE_RESTORE_STATE = "restore-state"
STORE_RECT_DATA = "rect-data"
STORE_BOUNDING_BOX = "bounding-box"
STORE_TRIGGER_GEN_DEPTHMAP = "trigger-generate-depthmap"
STORE_TRIGGER_UPDATE_DEPTHMAP = "trigger-update-depthmap"
STORE_UPDATE_THRESHOLD_CONTAINER = "update-thresholds-container"
STORE_CLICKED_POINT = "clicked-point"
STORE_CLEAR_PREVIEW = "clear-preview"

TEXT_POSITIVE_PROMPT = "positive-prompt"
TEXT_NEGATIVE_PROMPT = "negative-prompt"

CTR_INPUT_IMAGE = "input-image-container"
IMAGE = "image"
CANVAS = "canvas"
PREVIEW_CANVAS = "preview-canvas"
UPLOAD_IMAGE = "upload-image"
UPLOAD_STATE = "upload-state"
UPLOAD_SLICE = "slice-upload"
BTN_SAVE_STATE = "save-state"

SLIDER_INPAINT_STRENGTH = "inpaint-strength"
SLIDER_INPAINT_GUIDANCE = "inpaint-guidance"
SLIDER_NUM_SLICES = "num-slices-slider"
SLIDER_MASK_PADDING = "mask-padding-slider"
SLIDER_MASK_BLUR = "mask-blur-slider"

SLIDER_CAMERA_DISTANCE = "camera-distance-slider"
SLIDER_MAX_DISTANCE = "max-distance-slider"
SLIDER_FOCAL_LENGTH = "focal-length-slider"
SLIDER_DISPLACEMENT = "displacement-slider"
SLIDER_NUM_FRAMES = "number-of-frames-slider"

DROPDOWN_DEPTH_MODEL = "depth-model-dropdown"
DROPDOWN_INPAINT_MODEL = "inpainting-model-dropdown"
DROPDOWN_MODE_SELECTOR = "mode-selector"

INPUT_EXTERNAL_SERVER = "external-server-address"
BTN_EXTERNAL_TEST_CONNECTION = "external-test-connection-button"

UPLOAD_COMFYUI_WORKFLOW = "comfyui-workflow-upload"
CTR_COMFYUI_WORKFLOW = "comfyui-workflow-container"
CTR_AUTOMATIC_CONFIG = "automatic-config-container"

INPUT_API_KEY = "api-key"
BTN_VALIDATE_API_KEY = "validate-api-key"
CTR_API_KEY = "api-key-container"

DOWNLOAD_IMAGE = "download-image"
DOWNLOAD_GLTF = "download-gltf"
DOWNLOAD_ANIMATION = "download-animation"

BTN_EXPORT_ANIMATION = "animation-export"

CTR_THRESHOLDS = "thresholds-container"
CTR_SLICE_IMAGES = "slice-img-container"

BTN_GENERATE_INPAINTING = "generate-inpainting-button"
BTN_FILL_INPAINTING = "fill-inpainting-button"
BTN_ENHANCE = "enhance-button"
BTN_ERASE_INPAINTING = "erase-inpainting-button"
BTN_APPLY_INPAINTING = "apply-inpainting-button"
CTR_INPAINTING_OUTPUT = "inpainting-output"
CTR_INPAINTING_IMAGES = "inpainting-img-container"
CTR_INPAINTING_DISPLAY = "inpainting-image-display"
CTR_INPAINTING_COLUNM = "inpainting-column"

ID_INPAINTING_IMAGE = "inpainting-image"

ID_SLICE_DEPTH_DISPLAY = "depth-display"
ID_SLICE_OVERLAY = "slicer-overlay"
INPUT_SLICE_DEPTH = "depth-input"

BTN_GENERATE_SLICE = "generate-slice-button"
BTN_BALANCE_SLICE = "balance-slice-button"
BTN_CREATE_SLICE = "create-slice-button"
BTN_DELETE_SLICE = "delete-slice-button"
BTN_ADD_SLICE = "add-to-slice-button"
BTN_REMOVE_SLICE = "remove-from-slice-button"
BTN_COPY_SLICE = "copy-button"
BTN_PASTE_SLICE = "paste-button"

BTN_CLEAR_CANVAS = "clear-canvas"
BTN_ERASE_MODE = "erase-mode-canvas"
BTN_LOAD_CANVAS = "load-canvas"

CTR_CANVAS_BUTTONS = "canvas-buttons"
NAV_ZOOM_OUT = "nav-zoom-out"
NAV_UP = "nav-up"
NAV_DOWN = "nav-down"
NAV_LEFT = "nav-left"
NAV_RIGHT = "nav-right"
NAV_ZOOM_IN = "nav-zoom-in"
NAV_RESET = "nav-reset"

CTR_SEG_BUTTONS = "segmentation-buttons"
SEG_TOGGLE_CHECKERBOARD = "toggle-checkerboard"
SEG_INVERT_MASK = "invert-mask"
SEG_FEATHER_MASK = "feather-mask"
SEG_MULTI_POINT = "multi-point"
SEG_MULTI_COMMIT = "multi-commit"

CTR_PROGRESS_BAR = "progress-bar-container"
PROGRESS_INTERVAL = "progress-interval"

CTR_GLTF_OUTPUT = "gen-gltf-output"
BTN_GLTF_CREATE = "gltf-create"
BTN_GLTF_EXPORT = "gltf-export"
BTN_UPSCALE_TEXTURES = "upscale-textures"

CTR_MODEL_VIEWER = "model-viewer-container"
IFRAME_MODEL_VIEWER = "model-viewer"

CTR_HELP_WINDOW = "help-window"
STORE_CURRENT_TAB = "current-tab"

CHECKLIST_DOF = "toggle-dof-support"
CHECKLIST_REGION_OF_INTEREST = "toggle-region-of-interest"
